Package: inlpubs
Title: USGS INL Project Office Publications
Version: 1.4.0
Authors@R: c(
    person(given = c("Jason", "C."),
           family = "Fisher",
           role = c("aut", "cre"),
           email = "jfisher@usgs.gov",
           comment = c(ORCID = "0000-0001-9032-8912")),
    person(given = c("Kerri", "C."),
           family = "Treinen",
           role = "aut",
           email = "ktreinen@usgs.gov",
           comment = c(ORCID = "0000-0003-0645-6810")),
    person(given = c("Allison", "R."),
           family = "Trcka",
           role = "aut",
           email = "atrcka@usgs.gov",
           comment = c(ORCID = "0000-0001-8498-4737")))
Description: Provides bibliographic information and term-frequency text analysis
    tools for publications of the U.S. Geological Survey (USGS) Idaho National
    Laboratory (INL) Project Office. Includes datasets of publications, authors,
    and term frequencies, along with functions to search terms, build word
    clouds, and extract text and cover images from publication documents.
Depends: R (>= 4.1)
Imports: checkmate, knitr, kableExtra, parallel, stats, stringi, tm,
        utils
Suggests: callr, chromote, cli, connectapi, covr, cyclocomp, graphics,
        htmltools, htmlwidgets, jsonlite, lintr, magick, markdown,
        pdftools (>= 2.0), pkgbuild, pkgdown, pkgload, png, rcmdcheck,
        reactable, renv, rmarkdown, rsconnect, RWeka, tesseract,
        textutils, tinytest, webshot2, wordcloud2
License: CC0
URL: https://rconnect.usgs.gov/INLPO/inlpubs-main/,
        https://code.usgs.gov/inl/inlpubs
BugReports: https://code.usgs.gov/inl/inlpubs/-/issues
Copyright: This software is in the public domain because it contains
        materials that originally came from the United States
        Geological Survey (USGS), an agency of the United States
        Department of Interior. For more information, see the official
        USGS copyright policy at
        https://www.usgs.gov/information-policies-and-instructions/copyrights-and-credits
Encoding: UTF-8
SystemRequirements: Complete functionality necessitates Amazon Corretto
        (win), and default-jre, pandoc, libxml2-dev,
        libpoppler-cpp-dev, libmagick++-dev, optipng, libtesseract-dev,
        libleptonica-dev, tesseract-ocr-eng (deb)
LazyData: true
LazyDataCompression: xz
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-07-06 22:02:00 UTC; jfisher
Author: Jason C. Fisher [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-9032-8912>),
  Kerri C. Treinen [aut] (ORCID: <https://orcid.org/0000-0003-0645-6810>),
  Allison R. Trcka [aut] (ORCID: <https://orcid.org/0000-0001-8498-4737>)
Maintainer: Jason C. Fisher <jfisher@usgs.gov>
Repository: CRAN
Date/Publication: 2026-07-06 22:40:02 UTC
Built: R 4.6.0; ; 2026-07-06 23:43:41 UTC; unix

# Generated by roxygen2: do not edit by hand

export(add_content)
export(filter_data_list)
export(get_pdf_image)
export(get_pdf_text)
export(get_person)
export(make_webpages)
export(make_wordcloud)
export(mine_text)
export(plot_pubs)
export(search_terms)

# inlpubs <img src="man/figures/logo.svg" alt="inlpubs" align="right" width="152px" />

[![cran-version-image](https://www.r-pkg.org/badges/version/inlpubs)](https://CRAN.R-project.org/package=inlpubs)
[![pipeline-status-image](https://code.usgs.gov/inl/inlpubs/badges/main/pipeline.svg)](https://code.usgs.gov/inl/inlpubs/-/commits/main)
[![coverage-status-image](https://code.usgs.gov/inl/inlpubs/badges/main/coverage.svg)](https://code.usgs.gov/inl/inlpubs/-/commits/main)

## Description

The **inlpubs** [R](https://www.r-project.org/) package enables users to
search and analyze 396 publications spanning the 77-year history of the
U.S. Geological Survey (USGS) Idaho Water Science Center and the Idaho
National Laboratory Project Office
([INLPO](https://www.usgs.gov/centers/idaho-water-science-center/science/idaho-national-laboratory-project-office)).
Authored by 267 researchers, these publications provide valuable
insights into the impacts of waste disposal on the eastern Snake River
Plain aquifer and the availability of water for long-term consumptive
and industrial use. The information within these publications is
critical for the management and utilization of the aquifer by the Idaho
National Laboratory ([INL](https://inl.gov/)) and the State of Idaho.
USGS geohydrologic studies and monitoring, initiated in 1920, were
conducted in cooperation with the U.S. Department of Energy Idaho
Operations Office.

<p>

<figure class="figure">

<img src="man/figures/wordcloud.png" class="figure-img img-fluid" alt="photo" width="860px" />
<figcaption class="figure-caption">

A word cloud showcasing the 200 most frequently used words in titles and
abstracts.
</figcaption>

</figure>

</p>

## Installation

To install the current release of the package from
[CRAN](https://CRAN.R-project.org/package=inlpubs), you can use the
following command in R:

``` r
install.packages("inlpubs")
```

To install the package along with its dependencies, which are required
to run examples in the package help documentation and build package
datasets, run:

``` r
install.packages("inlpubs", dependencies = TRUE)
```

To install the development version of the package, you can either clone
the repository and build it from source, or run the following commands.
Make sure the **remotes** package for R is installed.

``` r
remotes::install_gitlab(
  repo = "inl/inlpubs@develop",
  host = "code.usgs.gov",
  dependencies = TRUE
)
```

## Usage

Examples are given in the package help pages. To access these documents,
run:

``` r
library("inlpubs")
help(package = "inlpubs")
```

## Test

The **tinytest** package is used for unit testing. To run the tests, use
the following command:

``` r
tinytest::test_package("inlpubs")
```

## Authors

- Jason C. Fisher (ORCID iD
  [0000-0001-9032-8912](https://orcid.org/0000-0001-9032-8912))
- Kerri C. Treinen (ORCID iD
  [0000-0003-0645-6810](https://orcid.org/0000-0003-0645-6810))
- Allison R. Trcka (ORCID iD
  [0000-0001-8498-4737](https://orcid.org/0000-0001-8498-4737))

## Point of Contact

Jason C. Fisher (<jfisher@usgs.gov>)

## Suggested Citation

To cite **inlpubs** in publications, please use:

Fisher, J.C., 2022, inlpubs—Bibliographic information for the U.S.
Geological Survey Idaho National Laboratory Project Office: U.S.
Geological Survey software release, R package, Reston, Va.,
<https://doi.org/10.5066/P9I3GWWU>.

## Contributing

We value your contributions and suggestions on how to make these
materials more useful to the community. Please feel free to share your
thoughts by commenting on the [issue
tracker](https://code.usgs.gov/inl/inlpubs/-/issues) or opening a [merge
request](https://code.usgs.gov/inl/inlpubs/-/merge_requests) to
contribute.

## Code of Conduct

All contributions to- and interactions surrounding- this project will
abide by the [USGS Code of Scientific
Conduct](https://www.usgs.gov/office-of-science-quality-and-integrity/fundamental-science-practices).

<!-- Embedded References -->

## Disclaimer

This software is preliminary or provisional and is subject to revision.
It is being provided to meet the need for timely best science. The
software has not received final approval by the U.S. Geological Survey
(USGS). No warranty, expressed or implied, is made by the USGS or the
U.S. Government as to the functionality of the software and related
material nor shall the fact of release constitute any such warranty. The
software is provided on the condition that neither the USGS nor the U.S.
Government shall be held liable for any damages resulting from the
authorized or unauthorized use of the software.

Any use of trade, product, or firm names is for descriptive purposes
only and does not imply endorsement by the U.S. Government.

## License

Unless otherwise noted, this project is in the public domain in the
United States because it contains materials that originally came from
the United States Geological Survey, an agency of the United States
Department of Interior. For more information, see the official USGS
copyright policy at
[copyrights-and-credits](https://www.usgs.gov/information-policies-and-instructions/copyrights-and-credits).

Additionally, we waive copyright and related rights in the work
worldwide through the CC0 1.0 Universal public domain dedication.

#### CC0 1.0 Universal Summary

This is a human-readable summary of the [Legal Code (read the full
text)](https://creativecommons.org/publicdomain/zero/1.0/legalcode).

##### No Copyright

The person who associated a work with this deed has dedicated the work
to the public domain by waiving all of his or her rights to the work
worldwide under copyright law, including all related and neighboring
rights, to the extent allowed by law.

You can copy, modify, distribute and perform the work, even for
commercial purposes, all without asking permission.

##### Other Information

In no way are the patent or trademark rights of any person affected by
CC0, nor are the rights that other persons may have in the work or in
how the work is used, such as publicity or privacy rights.

Unless expressly stated otherwise, the person who associated a work with
this deed makes no warranties about the work, and disclaims liability
for all uses of the work, to the fullest extent permitted by applicable
law. When using or citing the work, you should not imply endorsement by
the author or the affirmer.

<!-- Embedded References -->

## Support

The Idaho National Laboratory Project Office of the USGS supports the
development and maintenance of **inlpubs**. Resources are available
primarily for maintenance and responding to user questions, and the
development team determines the priorities for the development of new
features.

## Additional Publication Details

Additional metadata about this publication, not found in other parts of
the page is in this table.

<!--html_preserve-->

<table>

<tbody>

<tr>

<th scope="row">

Publication type
</th>

<td>

Formal R language package
</td>

</tr>

<tr>

<th scope="row">

DOI
</th>

<td>

10.5066/P9I3GWWU
</td>

</tr>

<tr>

<th scope="row">

Year published
</th>

<td>

2022
</td>

</tr>

<tr>

<th scope="row">

Version
</th>

<td>

<a href='https://code.usgs.gov/inl/inlpubs/-/tree/v1.0.0'>1.0.0</a>
</td>

</tr>

<tr>

<th scope="row">

IPDS
</th>

<td>

IP-130851
</td>

</tr>

</tbody>

</table>

<cr><!--/html_preserve-->

<!-- Embedded References -->

# inlpubs <img src="man/figures/logo.svg" alt="inlpubs" align="right" width="152px" />

[![cran-version-image](https://www.r-pkg.org/badges/version/inlpubs)](https://CRAN.R-project.org/package=inlpubs)
[![pipeline-status-image](https://code.usgs.gov/inl/inlpubs/badges/main/pipeline.svg)](https://code.usgs.gov/inl/inlpubs/-/commits/main)
[![coverage-status-image](https://code.usgs.gov/inl/inlpubs/badges/main/coverage.svg)](https://code.usgs.gov/inl/inlpubs/-/commits/main)

The **inlpubs** [R](https://www.r-project.org/) package provides tools
to search and analyze 399
[publications](https://rconnect.usgs.gov/INLPO/inlpubs-main/articles/inlpo-pubs.html)
spanning 78 years of work by the U.S. Geological Survey (USGS) Idaho
Water Science Center and the Idaho National Laboratory Project Office
([INLPO](https://www.usgs.gov/centers/idaho-water-science-center/science/idaho-national-laboratory-project-office)).
Written by 267 researchers, these publications document the effects of
waste disposal on the eastern Snake River Plain aquifer and assess the
availability of water for long-term consumptive and industrial use.
Their findings are central to how the Idaho National Laboratory
([INL](https://inl.gov/)) and the State of Idaho manage and use the
aquifer. The underlying USGS geohydrologic studies and monitoring began
in 1920 and have been carried out in cooperation with the U.S.
Department of Energy Idaho Operations Office.

<p>

<figure class="figure">

<img src="man/figures/wordcloud.png" class="figure-img img-fluid" alt="Word cloud of the 200 most frequent words in publication titles and abstracts." width="860px" />
<figcaption class="figure-caption">

A word cloud showcasing the 200 most frequently used words in titles and
abstracts.
</figcaption>

</figure>

</p>

## Installation

To install the current release from
[CRAN](https://CRAN.R-project.org/package=inlpubs), run:

``` r
install.packages("inlpubs")
```

To install the package together with the dependencies needed to run the
help-page examples and to build the package datasets, run:

``` r
install.packages("inlpubs", dependencies = TRUE)
```

To install the development version, either clone the repository and
build it from source, or run the following (the **remotes** package must
be installed first):

``` r
remotes::install_gitlab(
  repo = "inl/inlpubs@develop",
  host = "code.usgs.gov",
  dependencies = TRUE
)
```

## Usage

Worked examples are provided throughout the package help pages. To
browse them, run:

``` r
library("inlpubs")
help(package = "inlpubs")
```

## Test

Unit tests are written with the **tinytest** package. To run them, use:

``` r
tinytest::test_package("inlpubs")
```

## Authors

- Jason C. Fisher (ORCID iD
  [0000-0001-9032-8912](https://orcid.org/0000-0001-9032-8912))
- Kerri C. Treinen (ORCID iD
  [0000-0003-0645-6810](https://orcid.org/0000-0003-0645-6810))
- Allison R. Trcka (ORCID iD
  [0000-0001-8498-4737](https://orcid.org/0000-0001-8498-4737))

## Point of Contact

Jason C. Fisher (<jfisher@usgs.gov>)

## Suggested Citation

To cite **inlpubs** in publications, please use:

Fisher, J.C., 2022, inlpubs—Bibliographic information for the U.S.
Geological Survey Idaho National Laboratory Project Office: U.S.
Geological Survey software release, R package, Reston, Va.,
<https://doi.org/10.5066/P9I3GWWU>.

## Contributing

We welcome contributions and suggestions for making these materials more
useful to the community. Please share your ideas on the [issue
tracker](https://code.usgs.gov/inl/inlpubs/-/issues), or open a [merge
request](https://code.usgs.gov/inl/inlpubs/-/merge_requests) to
contribute directly.

## Code of Conduct

All contributions to- and interactions surrounding- this project will
abide by the [USGS Code of Scientific
Conduct](https://www.usgs.gov/office-of-science-quality-and-integrity/fundamental-science-practices).

<!-- Embedded References -->

## Disclaimer

This software is preliminary or provisional and is subject to revision.
It is being provided to meet the need for timely best science. The
software has not received final approval by the U.S. Geological Survey
(USGS). No warranty, expressed or implied, is made by the USGS or the
U.S. Government as to the functionality of the software and related
material nor shall the fact of release constitute any such warranty. The
software is provided on the condition that neither the USGS nor the U.S.
Government shall be held liable for any damages resulting from the
authorized or unauthorized use of the software.

Any use of trade, product, or firm names is for descriptive purposes
only and does not imply endorsement by the U.S. Government.

## License

Unless otherwise noted, this project is in the public domain in the
United States because it contains materials that originally came from
the United States Geological Survey, an agency of the United States
Department of Interior. For more information, see the official USGS
copyright policy at
[copyrights-and-credits](https://www.usgs.gov/information-policies-and-instructions/copyrights-and-credits).

Additionally, we waive copyright and related rights in the work
worldwide through the CC0 1.0 Universal public domain dedication.

#### CC0 1.0 Universal Summary

This is a human-readable summary of the [Legal Code (read the full
text)](https://creativecommons.org/publicdomain/zero/1.0/legalcode).

##### No Copyright

The person who associated a work with this deed has dedicated the work
to the public domain by waiving all of his or her rights to the work
worldwide under copyright law, including all related and neighboring
rights, to the extent allowed by law.

You can copy, modify, distribute and perform the work, even for
commercial purposes, all without asking permission.

##### Other Information

In no way are the patent or trademark rights of any person affected by
CC0, nor are the rights that other persons may have in the work or in
how the work is used, such as publicity or privacy rights.

Unless expressly stated otherwise, the person who associated a work with
this deed makes no warranties about the work, and disclaims liability
for all uses of the work, to the fullest extent permitted by applicable
law. When using or citing the work, you should not imply endorsement by
the author or the affirmer.

<!-- Embedded References -->

## Support

The development and maintenance of **inlpubs** are supported by the
Idaho National Laboratory Project Office of the USGS. Resources are
directed primarily toward maintenance and responding to user questions;
the development team sets the priorities for new features.

## Additional Publication Details

The table below lists additional metadata about this publication that
does not appear elsewhere on this page.

<!--html_preserve-->

<table>

<tbody>

<tr>

<th scope="row">

Publication type
</th>

<td>

Formal R language package
</td>

</tr>

<tr>

<th scope="row">

DOI
</th>

<td>

10.5066/P9I3GWWU
</td>

</tr>

<tr>

<th scope="row">

Year published
</th>

<td>

2022
</td>

</tr>

<tr>

<th scope="row">

Version
</th>

<td>

<a href='https://code.usgs.gov/inl/inlpubs/-/tree/v1.0.0'>1.0.0</a>
</td>

</tr>

<tr>

<th scope="row">

IPDS
</th>

<td>

IP-130851
</td>

</tr>

</tbody>

</table>

<cr><!--/html_preserve-->

<!-- Embedded References -->

Logo image files (man/figures/logo.ai, man/figures/logo.svg, and pkgdown/favicon/*.png) are a derivative of the [Documents_icon_-_noun_project_5020.svg] file, accessed on 2023-08-14.

* **File URL:** <https://upload.wikimedia.org/wikipedia/commons/6/68/Documents_icon_-_noun_project_5020.svg>

* **Attribution:** Dmitry Baranovskiy from The Noun Project, Creative Commons Attribution 3.0 Unported <https://creativecommons.org/licenses/by/3.0/deed.en> license.

[Documents_icon_-_noun_project_5020.svg]: https://commons.wikimedia.org/wiki/File:Documents_icon_-_noun_project_5020.svg

# setup tinytest for checkmate functionality
library("tinytest")
library("checkmate")
using("checkmate")

# test for missing annotation sources
is <- pubs$annotation_src[!is.na(pubs$annotation)] |> anyNA()
expect_false(is)

#' Select Persons by Identifier
#'
#' @description Select one or more persons from a named list by their unique
#'   identifiers, preserving the requested order. All identifiers must be present;
#'   any that are missing raise an error.
#'
#' @param x 'character' vector.
#'   Identifier(s) of the person(s) to select. Must be non-missing and unique, and
#'   each must match a name in `persons`.
#'
#' @param persons 'person' named list.
#'   Information about an arbitrary number of persons, as produced by
#'   [utils::person()]. Each element is named with the identifier that uniquely
#'   identifies that person.
#'
#' @return A 'person' list containing the requested subset of `persons`, in the
#'   order given by `x`.
#'
#' @author J.C. Fisher, U.S. Geological Survey, Idaho Water Science Center
#'
#' @export
#'
#' @keywords internal
#'
#' @examples
#' get_person("jfisher", persons = inlpubs::authors$person)

get_person <- function(x, persons) {
  # check arguments
  checkmate::assert_character(x, any.missing = FALSE, unique = TRUE, .var.name = "x")
  checkmate::assert_class(persons, classes = "person")

  # require that every identifier is present
  missing_id <- !(x %in% names(persons))
  if (any(missing_id)) {
    txt <- x[missing_id] |> sQuote(q = FALSE) |> paste(collapse = ", ")
    stop("Missing entry for person identifier(s): ", txt, call. = FALSE)
  }

  persons[x]
}

# setup tinytest for checkmate functionality
library("tinytest")
library("checkmate")
using("checkmate")

# ------ get_person ------

persons <- authors[["person"]]

# returns a person object
p <- get_person("jfisher", persons = persons)
expect_true(inherits(p, "person"))

# returns the correct number of entries
expect_equal(length(p), 1L)

# multiple ids returned in requested order
ids <- c("jfisher", "rbartholomay")
p2 <- get_person(ids, persons = persons)
expect_equal(length(p2), 2L)
expect_equal(names(p2), ids)

# order is preserved when reversed
p3 <- get_person(rev(ids), persons = persons)
expect_equal(names(p3), rev(ids))

# missing id raises an error
expect_error(get_person("zzznobody", persons = persons))

# duplicate ids raise an error
expect_error(get_person(c("jfisher", "jfisher"), persons = persons))

# NA raises an error
expect_error(get_person(NA_character_, persons = persons))

# ------ filter_data_list ------

if (requireNamespace("htmltools", quietly = TRUE)) {
  f <- filter_data_list("my-table")

  # returns a function
  expect_true(is.function(f))

  # calling the function produces an htmltools tagList
  result <- f(c("b", "a", "c"), "type")
  expect_true(inherits(result, "shiny.tag.list"))

  # datalist id is constructed from table_id and column name
  html <- as.character(htmltools::tagList(result))
  expect_true(grepl("my-table-type-list", html))

  # options are sorted (values appear as attributes)
  a_pos <- regexpr('value="a"', html, fixed = TRUE)
  b_pos <- regexpr('value="b"', html, fixed = TRUE)
  expect_true(a_pos < b_pos)

  # aria-label is set
  expect_true(grepl('aria-label="Filter type"', html))

  # custom style is applied
  f_styled <- filter_data_list("t", style = "color:red;")
  result_styled <- f_styled("x", "col")
  html_styled <- as.character(htmltools::tagList(result_styled))
  expect_true(grepl("color:red;", html_styled))

  # invalid table_id raises an error
  expect_error(filter_data_list(""))
  expect_error(filter_data_list(123))
}

#' Create a Word Cloud from Term Frequencies
#'
#' @description Create a word cloud from a term-frequency table and save it as a PNG
#'   file. The most frequent terms are rendered, laid out within the chosen shape.
#'   Requires the \pkg{htmltools}, \pkg{htmlwidgets}, \pkg{magick}, \pkg{webshot2},
#'   and \pkg{wordcloud2} packages. System dependencies are
#'   [ImageMagick](https://imagemagick.org/) (for displaying the PNG),
#'   [OptiPNG](https://optipng.sourceforge.net/) (for PNG compression), and a
#'   [Chrome](https://www.google.com/chrome/) or Chromium-based browser with
#'   support for the Chrome DevTools protocol. Use
#'   [chromote::find_chrome()] to locate the browser.
#'
#' @param x 'data.frame'.
#'   Term-frequency table containing at least the columns `term` and `freq`.
#'
#' @param max_terms 'integer' number.
#'   Maximum number of terms to include, taken in order of decreasing frequency.
#'   Defaults to 200.
#'
#' @param size 'numeric' number.
#'   Font-size scaling factor. Defaults to 1.
#'
#' @param shape 'character' string.
#'   Shape of the cloud. One of `"circle"` (the default), `"cardioid"`, `"diamond"`,
#'   `"triangle-forward"`, `"triangle"`, `"pentagon"`, or `"star"`.
#'
#' @param ellipticity 'numeric' number between 0 and 1.
#'   Degree of flattening applied to the shape. Defaults to 0.65.
#'
#' @param ...
#'   Additional arguments passed to [wordcloud2::wordcloud2()].
#'
#' @param width 'integer' number.
#'   Image width in pixels. Defaults to 910. Image height is derived from `width`
#'   and `ellipticity`.
#'
#' @param output 'character' string.
#'   Path to the output PNG file. Defaults to a temporary file.
#'
#' @param display 'logical' flag.
#'   Whether to display the saved PNG in a graphics window. Requires the \pkg{magick}
#'   package. Defaults to `FALSE`.
#'
#' @return Invisibly returns the file path to the word-cloud PNG.
#'
#' @author J.C. Fisher, U.S. Geological Survey, Idaho Water Science Center
#'
#' @export
#'
#' @seealso [`mine_text`] to build a term-frequency table.
#'
#' @examples
#' \dontrun{
#'   d <- wordcloud2::demoFreq |> head(n = 10)
#'   colnames(d) <- c("term", "freq")
#'   file <- make_wordcloud(d, display = interactive())
#'
#'   unlink(file)
#' }

make_wordcloud <- function(
  x,
  max_terms = 200,
  size = 1,
  shape = "circle",
  ellipticity = 0.65,
  ...,
  width = 910,
  output = NULL,
  display = FALSE
) {
  # check required packages
  pkgs <- c(
    "htmltools",
    "htmlwidgets",
    "magick",
    "utils",
    "webshot2",
    "wordcloud2"
  )
  installed <- vapply(pkgs, FUN = requireNamespace, FUN.VALUE = logical(1), quietly = TRUE)
  if (any(!installed)) {
    txt <- pkgs[!installed] |> sQuote(q = FALSE) |> paste(collapse = ", ")
    stop("Word cloud requires missing package(s): ", txt, call. = FALSE)
  }

  # check arguments
  checkmate::assert_data_frame(
    x,
    types = c("factor", "character", "integerish"),
    any.missing = FALSE,
    min.rows = 3,
    min.cols = 2
  )
  checkmate::assert_names(colnames(x), must.include = c("term", "freq"))
  checkmate::assert_count(max_terms, positive = TRUE)
  checkmate::assert_number(size, lower = 0, finite = TRUE)
  shape <- match.arg(
    shape,
    choices = c(
      "circle",
      "cardioid",
      "diamond",
      "triangle-forward",
      "triangle",
      "pentagon",
      "star"
    )
  )
  checkmate::assert_number(ellipticity, lower = 0, upper = 1, finite = TRUE)
  checkmate::assert_count(width, positive = TRUE)
  checkmate::assert_flag(display)
  if (is.null(output)) {
    output <- tempfile(fileext = ".png")
  }
  output <- normalizePath(output, winslash = "/", mustWork = FALSE)
  checkmate::assert_path_for_output(output, overwrite = TRUE, extension = "png")

  # keep the most frequent terms
  d <- x[order(x$freq, decreasing = TRUE), ]
  d <- utils::head(d, n = max_terms)

  # create the word-cloud HTML widget
  wc <- wordcloud2::wordcloud2(d, size = size, shape = shape, ellipticity = ellipticity, ...)

  # hide hover labels
  sty <- htmltools::HTML(".wcLabel {display: none;}")
  tag <- htmltools::tags$style(sty) |> htmltools::tags$head()
  wc <- htmlwidgets::prependContent(wc, tag)

  # save the widget to a temporary HTML file
  html <- tempfile(fileext = ".html")
  htmlwidgets::saveWidget(wc, html, selfcontained = FALSE)

  # ensure HTML artifacts are removed even if the screenshot step errors
  on.exit(
    unlink(
      list.files(
        path = dirname(html),
        pattern = sub("\\.html$", "", basename(html)),
        full.names = TRUE
      ),
      recursive = TRUE
    ),
    add = TRUE
  )

  # screenshot the HTML widget into the PNG file
  webshot2::webshot(
    url = html,
    file = output,
    vwidth = width,
    vheight = as.integer(width * ellipticity),
    cliprect = "viewport",
    delay = 10
  )

  # recompress the PNG
  suppressWarnings(
    system2(
      command = "optipng",
      args = c("-quiet", "-strip all", "-o7", shQuote(output))
    )
  )

  # optionally display the result
  if (display) {
    magick::image_read(output) |> plot()
  }

  invisible(output)
}

#' Add Content from PDF Documents
#'
#' @description Extract the text or cover image from publication PDF documents and
#'   save the result to disk, for incorporation into the \pkg{inlpubs} package corpus.
#'   Publications are selected by identifier or by year, and each must already be
#'   present in the `pubs` dataset before this function is run.
#'
#' @details For each selected publication, the function locates its primary PDF (the
#'   first file listed in the publication's `files` entry) within the source directory,
#'   then delegates extraction to [`get_pdf_text`] (for `type = "text"`) or
#'   [`get_pdf_image`] (for `type = "image"`). Extra arguments passed through `...`
#'   are forwarded to whichever of those functions is used.
#'
#'   Source documents are expected to follow the archive layout
#'   `srcdir/<year>/<pub_id>/<file>`, for example
#'   `archive/2005/KnobelOthers2005/ofr20051223.pdf`. Output files are written to
#'   `destdir` and named `pub-<pub_id>`, with the appropriate extension (`.txt` or
#'   `.jpg`) supplied by the extraction function.
#'
#'   Publications whose `files` entry is empty, and those listed in `ignore`, are
#'   skipped with a message. By default, existing output files are left in place;
#'   set `overwrite = TRUE` to regenerate them.
#'
#'   Extraction failures (including renderer crashes or timeouts on problematic
#'   PDFs) are caught per publication: the offending document is reported and
#'   skipped, and processing continues with the next one. The identifiers of any
#'   failed documents are returned in the `"failed"` attribute of the result.
#'
#'   Processing the full archive can take a considerable amount of time. For routine
#'   use, run this function interactively and one publication year at a time so that
#'   progress can be monitored.
#'
#' @param pub_id 'character' vector.
#'   Unique identifier(s) of the publication(s) to process.
#'   Publications may instead, or additionally, be selected with `year`.
#'
#' @param year 'integer' vector.
#'   Year(s) of publication. All publications from the given year(s) are selected,
#'   in addition to any specified by `pub_id`.
#'
#' @param type 'character' string.
#'   Type of content to extract from each PDF, either `"text"` (the default) or
#'   `"image"` (a cover image).
#'
#' @param ...
#'   Additional arguments passed to the extraction function:
#'   [`get_pdf_text`] when `type = "text"`, or [`get_pdf_image`] when `type = "image"`.
#'
#' @param srcdir 'character' string.
#'   Source directory containing the publication PDFs, organized into
#'   year subdirectories (see Details). Defaults to `"archive"` in the working directory.
#'
#' @param destdir 'character' string.
#'   Destination directory for the extracted output files. Created if it does not
#'   already exist. Defaults to the session's temporary directory.
#'
#' @param ignore 'character' vector.
#'   Publication identifier(s) to skip. Useful for documents that are known to fail
#'   automatic extraction.
#'
#' @param pubs 'pub' table.
#'   Publications of the INLPO; see the [`pubs`] dataset for the expected data format.
#'
#' @param overwrite 'logical' flag.
#'   Whether to overwrite an existing output file. Defaults to `FALSE`, in which case
#'   publications with an existing output are skipped.
#'
#' @return Invisibly returns a 'character' vector of paths to the saved output files,
#'   carrying a `"failed"` attribute with the identifiers of any publications whose
#'   extraction failed. Returns `NULL` (invisibly) when no publications remain to
#'   process.
#'
#' @author J.C. Fisher, U.S. Geological Survey, Idaho Water Science Center
#'
#' @export
#'
#' @seealso [`get_pdf_text`] and [`get_pdf_image`] for the underlying extraction
#'   functions, and the [`pubs`] dataset for publication metadata.
#'
#' @examples
#' \dontrun{
#'   # Extract text for a single publication into the corpus folder
#'   add_content("KnobelOthers2005", destdir = "data-raw/corpus")
#'
#'   # Extract text for every publication from a given year
#'   add_content(year = 2005, destdir = "data-raw/corpus")
#'
#'   # Extract cover images into the vignettes folder
#'   add_content(year = 2005, type = "image", destdir = "vignettes")
#'
#'   # Process several years, skipping a document that fails extraction
#'   add_content(year = 2003:2005, ignore = "Mazurek2004", destdir = "data-raw/corpus")
#' }

add_content <- function(
  pub_id,
  year,
  type = c("text", "image"),
  ...,
  srcdir = "archive",
  destdir = tempdir(),
  ignore = NULL,
  pubs = inlpubs::pubs,
  overwrite = FALSE
) {
  # check arguments
  if (missing(pub_id)) {
    pub_id <- NULL
  }
  if (!missing(year)) {
    checkmate::assert_integerish(year, lower = min(pubs$year), upper = max(pubs$year), any.missing = FALSE)
    ids <- pubs$pub_id[pubs$year %in% unique(year)]
    pub_id <- c(pub_id, ids) |> unique()
  }
  checkmate::assert_subset(pub_id, choices = pubs$pub_id, empty.ok = TRUE)
  type <- match.arg(type)
  checkmate::assert_directory_exists(srcdir, access = "r")
  checkmate::assert_string(destdir, min.chars = 1)
  checkmate::assert_subset(ignore, choices = pubs$pub_id, empty.ok = TRUE)
  checkmate::assert_class(pubs, classes = c("pub", "data.frame"), null.ok = TRUE)
  checkmate::assert_flag(overwrite)

  # set extraction method
  extract <- switch(type, "text" = get_pdf_text, "image" = get_pdf_image)

  # remove ignored publication identifiers
  pub_id <- pub_id[!pub_id %in% ignore]

  # account for no publications
  if (length(pub_id) == 0) {
    return(invisible(NULL))
  }

  # look up each publication's primary document (first file listed)
  idxs <- match(pub_id, pubs$pub_id)
  years <- pubs$year[idxs]
  docs <- vapply(
    pubs$files[idxs],
    FUN = function(x) if (length(x) == 0) NA_character_ else x[1],
    FUN.VALUE = character(1)
  )

  # drop publications with a missing document, keeping all vectors aligned
  if (any(missing_doc <- is.na(docs))) {
    n_missing <- sum(missing_doc)
    if (requireNamespace("cli", quietly = TRUE)) {
      cli::cli_alert_info(
        "Skipping {n_missing} publication{?s} with no archived document"
      )
    } else {
      message("Skipping ", n_missing, " publication(s) with no archived document")
    }
    pub_id <- pub_id[!missing_doc]
    years <- years[!missing_doc]
    docs <- docs[!missing_doc]
  }

  # account for no remaining publications
  if (length(pub_id) == 0) {
    return(invisible(NULL))
  }

  # build input (source PDF) and output paths
  inputs <- file.path(srcdir, years, pub_id, docs)
  outputs <- file.path(destdir, sprintf("pub-%s", pub_id))

  # create destination directory
  dir.create(destdir, showWarnings = FALSE, recursive = TRUE)

  # extract content for each publication. Extraction is wrapped so that a failure
  # on one document (e.g. a renderer crash or timeout caught by the extraction
  # function) is reported and skipped rather than aborting the whole run. Failed
  # publications yield NA here and are recorded for the returned "failed" attribute.
  results <- vapply(
    seq_along(inputs),
    FUN = function(i) {
      id <- sQuote(pub_id[i])
      message("Extract publication ", id)

      # skip when output already exists and overwrite is not requested
      if (!overwrite && checkmate::test_file_exists(outputs[i], access = "rw")) {
        message("Skipping publication ", id, " (output exists)")
        return(outputs[i])
      }

      # skip when the source PDF is not present
      if (!checkmate::test_file_exists(inputs[i], access = "r")) {
        message("Missing source PDF for publication ", id)
        return(NA_character_)
      }

      # extract, catching errors (including the renderer crash/timeout errors
      # that the extraction functions raise) so one bad document cannot abort
      # the run
      tryCatch(
        extract(input = inputs[i], output = outputs[i], ...),
        error = function(e) {
          message("Failed to extract publication ", id, ": ", conditionMessage(e))
          NA_character_
        }
      )
    },
    FUN.VALUE = character(1)
  )

  # separate successes from failures
  failed <- pub_id[is.na(results)]
  paths <- results[!is.na(results)]

  if (length(failed) > 0) {
    message(
      "Failed to extract ",
      length(failed),
      " publication(s): ",
      paste(sQuote(failed), collapse = ", ")
    )
  }

  attr(paths, "failed") <- failed
  invisible(paths)
}

#' Plot Annual Publication Counts
#'
#' @description Draw a stacked bar chart of INLPO publication counts by year,
#'   with USGS and non-USGS publications distinguished by color.
#'
#' @param pubs 'pub' data frame.
#'   Publications of the INLPO; see the [`pubs`] dataset for the expected data
#'   format. Defaults to `inlpubs::pubs`.
#'
#' @param author_id 'character' vector, or `NULL`.
#'   One or more author identifiers used to subset `pubs` to publications that
#'   any of those authors appear on before plotting. The default (`NULL`)
#'   includes all publications (no author filtering).
#'
#' @param full_range 'logical' flag.
#'   Whether the x-axis spans the year range of all publications in `pubs`
#'   (before any author filtering). If `TRUE` (the default), the timeline is
#'   fixed to the full dataset, making per-author charts directly comparable.
#'   If `FALSE`, the timeline is limited to the year range of the filtered
#'   publications. Has no effect when `author_id` is `NULL`.
#'
#' @param legend_loc 'character' string.
#'   Position of the legend, passed as the keyword location to
#'   [graphics::legend()]. One of `"topleft"` (the default), `"topright"`,
#'   `"bottomleft"`, `"bottomright"`, `"top"`, `"bottom"`, `"left"`,
#'   `"right"`, or `"center"`.
#'
#' @param colors 'character' vector of length 2.
#'   Fill colors for the two series, in the order `c("USGS", "Non-USGS")`.
#'   Defaults to steel blue and burnt orange.
#'
#' @param ...
#'   Additional arguments passed to [graphics::barplot()].
#'
#' @return Invisibly returns a numeric vector giving the x-axis coordinates of
#'   the bar midpoints, as returned by [graphics::barplot()].
#'
#' @author J.C. Fisher, U.S. Geological Survey, Idaho Water Science Center
#'
#' @export
#'
#' @seealso [`pubs`] for the publications dataset.
#'
#' @examples
#' plot_pubs()

plot_pubs <- function(
  pubs = inlpubs::pubs,
  author_id = NULL,
  full_range = TRUE,
  legend_loc = "topleft",
  colors = c("USGS" = "#4682B4", "Non-USGS" = "#B47846"),
  ...
) {
  # check arguments
  checkmate::assert_class(pubs, classes = c("pub", "data.frame"))
  checkmate::assert_character(author_id, any.missing = FALSE, min.len = 1, null.ok = TRUE)
  checkmate::assert_flag(full_range)
  checkmate::assert_choice(
    legend_loc,
    choices = c(
      "topleft",
      "topright",
      "bottomleft",
      "bottomright",
      "top",
      "bottom",
      "left",
      "right",
      "center"
    )
  )
  checkmate::assert_character(colors, len = 2, any.missing = FALSE)
  checkmate::assert_names(names(colors), identical.to = c("USGS", "Non-USGS"))

  # capture the full year range before filtering (for a fixed timeline)
  full_yrs <- range(pubs$year)

  # subset to publications any of the requested authors appear on (NULL keeps all)
  if (!is.null(author_id)) {
    is_match <- vapply(
      pubs$author_id,
      FUN = function(ids) any(author_id %in% ids),
      FUN.VALUE = logical(1)
    )
    pubs <- pubs[is_match, , drop = FALSE]
    if (nrow(pubs) == 0L) {
      stop("No publications found for the specified author_id(s).", call. = FALSE)
    }
  }

  # tally counts per year over a gap-free year sequence (empty years are zero)
  yr_range <- if (full_range) full_yrs else range(pubs$year)
  all_yrs <- yr_range[1]:yr_range[2]
  is_usgs <- pubs$institution %in% "U.S. Geological Survey"
  total_by_yr <- table(factor(pubs$year, levels = all_yrs))
  usgs_by_yr <- table(factor(pubs$year[is_usgs], levels = all_yrs))

  # stack USGS and non-USGS series into a 2-row matrix
  counts <- rbind(
    "USGS" = as.integer(usgs_by_yr),
    "Non-USGS" = as.integer(total_by_yr) - as.integer(usgs_by_yr)
  )
  y_breaks <- pretty(c(0, max(colSums(counts))))
  y_breaks <- as.integer(y_breaks[y_breaks %% 1 == 0])
  ylim <- range(y_breaks)

  # draw the stacked bars with custom axes (margins restored on exit)
  old_par <- graphics::par(mar = c(2.1, 3.1, 1.1, 0.1))
  on.exit(graphics::par(old_par), add = TRUE)
  mp <- graphics::barplot(
    height = counts,
    ylab = "Number of publications per year",
    col = colors,
    space = 0.2,
    border = NA,
    ylim = ylim,
    yaxt = "n",
    xaxt = "n",
    cex.lab = 0.7,
    mgp = c(2, 0.5, 0),
    ...
  )

  # label years at pretty breaks (clamped to range, falling back to all years)
  pretty_yrs <- pretty(all_yrs, n = 7)
  pretty_yrs <- pretty_yrs[pretty_yrs >= min(all_yrs) & pretty_yrs <= max(all_yrs)]
  label_yrs <- if (length(pretty_yrs)) pretty_yrs else all_yrs
  is_label <- all_yrs %in% label_yrs

  # axes: labeled bottom/left, mirrored unlabeled top/right
  graphics::axis(
    1,
    at = mp[is_label],
    labels = all_yrs[is_label],
    tck = 0.02,
    lwd = 0,
    lwd.ticks = 0.5,
    cex.axis = 0.7,
    mgp = c(2, 0.5, 0)
  )
  graphics::axis(
    2,
    at = y_breaks,
    las = 2,
    tck = 0.02,
    lwd = 0,
    lwd.ticks = 0.5,
    cex.axis = 0.7,
    mgp = c(2, 0.5, 0)
  )
  graphics::axis(
    3,
    at = mp[is_label],
    labels = NA,
    tck = 0.02,
    lwd = 0,
    lwd.ticks = 0.5
  )
  graphics::axis(
    4,
    at = y_breaks,
    labels = NA,
    las = 2,
    tck = 0.02,
    lwd = 0,
    lwd.ticks = 0.5
  )

  # legend (only series present) and plot frame
  present <- rowSums(counts) > 0
  graphics::legend(
    x = legend_loc,
    legend = names(colors)[present],
    fill = colors[present],
    border = NA,
    inset = c(0.02, 0.05),
    cex = 0.7,
    box.lty = 1,
    box.lwd = 0.5,
    xpd = NA,
    bg = "#FFFFFFE7"
  )
  graphics::box(lwd = 0.5)

  invisible(mp)
}

# setup tinytest for checkmate functionality
library("tinytest")
library("checkmate")
using("checkmate")

# ------ default call ------

# returns invisible numeric vector of bar midpoints
mp <- plot_pubs()
expect_true(is.numeric(mp))
expect_true(length(mp) > 0L)

# ------ author_id filtering ------

# single author
mp_author <- plot_pubs(author_id = "jfisher")
expect_true(is.numeric(mp_author))
expect_true(length(mp_author) > 0L)

# multiple authors
mp_multi <- plot_pubs(author_id = c("jfisher", "ktreinen"))
expect_true(is.numeric(mp_multi))

# ------ full_range ------

# full_range = FALSE limits x-axis to author's year range
mp_full  <- plot_pubs(author_id = "jfisher", full_range = TRUE)
mp_short <- plot_pubs(author_id = "jfisher", full_range = FALSE)
expect_true(length(mp_full) >= length(mp_short))

# ------ custom colors ------

mp_cols <- plot_pubs(colors = c("USGS" = "#FF0000", "Non-USGS" = "#0000FF"))
expect_true(is.numeric(mp_cols))

# ------ par() is restored after call ------

graphics::par(mar = c(5, 4, 4, 2))
plot_pubs()
expect_equal(graphics::par("mar"), c(5, 4, 4, 2))

# ------ invalid arguments ------

expect_error(plot_pubs(author_id = "zzznobody"))
expect_error(plot_pubs(colors = c("#FF0000", "#0000FF")))           # unnamed
expect_error(plot_pubs(colors = c("A" = "#FF0000", "B" = "#000"))) # wrong names
expect_error(plot_pubs(pubs = data.frame()))                        # wrong class

#' Filter Data List Column
#'
#' @description Create a column filter for a \pkg{reactable} table that pairs a text
#'   input with an HTML `datalist`, giving users autocomplete suggestions drawn from
#'   the column's unique values.
#'   Requires that the \pkg{htmltools} package is available.
#'
#' @details The returned function is intended for use as a column's `filterInput` in
#'   [reactable::colDef()]. When rendered, it produces a text box backed by a
#'   `datalist` of the sorted unique values in that column; typing filters the table
#'   via `Reactable.setFilter()`, while the suggestions assist exact matches.
#'
#' @param table_id 'character' string.
#'   Unique identifier of the table the filter belongs to. Must match the `elementId`
#'   given to the table so that `Reactable.setFilter()` can target it.
#'
#' @param style 'character' string.
#'   CSS applied to the text input element. Defaults to a full-width, 28px-tall box.
#'
#' @return Returns a function of two arguments, `values` (the column's values) and
#'   `name` (the column name), which returns an \pkg{htmltools} tag list for the
#'   filter input.
#'
#' @author J.C. Fisher, U.S. Geological Survey, Idaho Water Science Center
#'
#' @keywords internal
#' @export
#'
#' @keywords internal
#'
#' @references Adapted from the \pkg{reactable} "Custom Filtering" vignette
#'   (\url{https://github.com/glin/reactable/blob/HEAD/vignettes/custom-filtering.Rmd}),
#'   by Greg Lin and Tanner Linsley, 2019 (MIT License). Accessed 2024-05-30.
#'
#' @examples
#' f <- filter_data_list("table-id")

filter_data_list <- function(
  table_id,
  style = "width: 100%; height: 28px;"
) {
  # check system dependencies
  if (!requireNamespace("htmltools", quietly = TRUE)) {
    stop("Creating a data list column filter requires the 'htmltools' package", call. = FALSE)
  }

  # check arguments
  checkmate::assert_string(table_id, min.chars = 1)
  checkmate::assert_string(style)

  # return the filter-input function used by reactable::colDef(filterInput = ...)
  function(values, name) {
    list_id <- sprintf("%s-%s-list", table_id, name)
    htmltools::tagList(
      htmltools::tags$input(
        type = "text",
        list = list_id,
        oninput = sprintf(
          "Reactable.setFilter('%s', '%s', event.target.value || undefined)",
          table_id,
          name
        ),
        "aria-label" = sprintf("Filter %s", name),
        style = style
      ),
      htmltools::tags$datalist(
        id = list_id,
        lapply(
          sort(unique(values)),
          FUN = function(x) htmltools::tags$option(value = x)
        )
      )
    )
  }
}

#' Create Author and Publication Webpages
#'
#' @description Create a webpage for each author and each publication, listing the
#'   author's publications or the publication's details. Each webpage is written as
#'   an [R Markdown](https://rmarkdown.rstudio.com/) (`.Rmd`) file.
#'
#' @param authors 'author' data frame.
#'   Contributing authors to the INLPO publications; see the [`authors`] dataset for
#'   the expected data format. When `NULL`, author webpages are not written.
#'
#' @param pubs 'pub' data frame.
#'   Publications of the INLPO; see the [`pubs`] dataset for the expected data format.
#'   When `NULL`, publication webpages are not written.
#'
#' @param destdir 'character' string.
#'   Destination directory for the written files, with tilde-expansion performed.
#'   Created if it does not exist. Defaults to a temporary directory.
#'
#' @param coverdir 'character' string.
#'   Directory searched for publication cover images named `pub-<id>.jpg`.
#'   Defaults to `"vignettes"`.
#'
#' @param quiet 'logical' flag.
#'   Whether to suppress progress messages. Defaults to `FALSE`.
#'
#' @return Invisibly returns `NULL`. Called for its side effect of writing files.
#'
#' @author J.C. Fisher, U.S. Geological Survey, Idaho Water Science Center
#'
#' @export
#'
#' @keywords internal
#'
#' @examples
#' destdir <- tempfile("")
#' make_webpages(
#'   authors = inlpubs::authors,
#'   pubs = inlpubs::pubs,
#'   destdir = destdir,
#'   quiet = TRUE
#' )
#'
#' unlink(destdir, recursive = TRUE)

make_webpages <- function(
  authors = NULL,
  pubs = NULL,
  destdir = tempdir(),
  coverdir = "vignettes",
  quiet = FALSE
) {
  # check arguments
  checkmate::assert_class(authors, classes = c("author", "data.frame"), null.ok = TRUE)
  checkmate::assert_class(pubs, classes = c("pub", "data.frame"), null.ok = TRUE)
  checkmate::assert_string(destdir)
  destdir <- path.expand(destdir) |> normalizePath(winslash = "/", mustWork = FALSE)
  dir.create(destdir, recursive = TRUE, showWarnings = FALSE)
  checkmate::assert_path_for_output(destdir, overwrite = TRUE)
  checkmate::assert_string(coverdir)
  checkmate::assert_flag(quiet)

  # write author webpages
  if (!is.null(authors)) {
    make_author_webpages(authors, pubs, destdir, coverdir, quiet)
  }

  # write publication webpages
  if (!is.null(pubs)) {
    make_pub_webpages(pubs, destdir, coverdir, quiet)
  }

  invisible()
}


# Helper to assemble Rmd sections and write to disk ----
#
# Builds each file's content as a single string and writes it with writeLines(),
# replacing the sink()-based approach. This avoids leaking sink connections if an
# error occurs mid-build, and makes the output assembly explicit and testable.

write_rmd_files <- function(files, contents, quiet) {
  for (i in seq_along(files)) {
    if (!quiet) {
      message("Writing ", sQuote(basename(files[i]), q = FALSE))
    }
    writeLines(contents[i], con = files[i])
  }
  invisible()
}


# Helper to atomically replace a set of pages in the destination ----

replace_pages <- function(contents, names, pattern, destdir, quiet) {
  tmp <- tempfile(pattern = "")
  dir.create(tmp, showWarnings = FALSE)
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)

  files <- file.path(tmp, names)
  write_rmd_files(files, contents, quiet)

  # replace existing pages in the destination
  list.files(path = destdir, pattern = pattern, full.names = TRUE) |> unlink()
  file.copy(from = files, to = destdir) |> invisible()

  invisible()
}


# Function to write author webpages ----

make_author_webpages <- function(authors, pubs, destdir, coverdir, quiet) {
  # check arguments
  checkmate::assert_class(authors, classes = c("author", "data.frame"))
  checkmate::assert_path_for_output(destdir, overwrite = TRUE)
  checkmate::assert_string(coverdir)
  checkmate::assert_flag(quiet)

  n <- nrow(authors)

  # make yaml headers
  headers <- vapply(
    authors$person,
    FUN = function(x) {
      title <- sprintf("title: \"%s\"", format(x, include = c("given", "family")))
      paste("---", title, "---", sep = "\n")
    },
    FUN.VALUE = character(1)
  )

  # make ORCID and email icons
  icons <- vapply(
    authors$person,
    FUN = function(x) {
      orcid <- x$comment["ORCID"]
      # subscripting NULL/absent comment yields length-0, so guard on length
      # AND non-NA before building the link
      if (length(orcid) > 0 && !is.na(orcid)) {
        href <- paste0("https://orcid.org/", orcid)
        title <- "Open the author&#8217;s ORCID record in a new browser tab."
        img <- "<img alt='ORCID logo' src='orcid.svg' width='16' height='16' />"
        orcid <- sprintf(
          "<a href='%s' title='%s' target='_blank'>%s&nbsp;%s</a>",
          href,
          title,
          img,
          orcid
        )
      } else {
        orcid <- NULL
      }
      email <- x$email
      # a missing email is character(0), so test with length()
      if (length(email) > 0 && nzchar(email)) {
        href <- paste0("mailto:", email)
        title <- "Contact the author"
        img <- "<img alt='Email logo' src='email.svg' width='16' height='16' />"
        email <- sprintf("<a href='%s' title='%s'>%s&nbsp;%s</a>", href, title, img, email)
      } else {
        email <- NULL
      }
      sprintf("<span>%s</span>", paste(c(orcid, email), collapse = "&nbsp;&nbsp;&nbsp;&nbsp;"))
    },
    FUN.VALUE = character(1)
  )

  # make counts table
  row_names <- c(
    "total_pub" = "Total publications",
    "single_authored" = "Single authored",
    "multi_authored" = "Multi authored",
    "first_authored" = "First authored"
  )
  # iterate by row index rather than apply() over a data frame with list-columns,
  # which is clearer and avoids coercion surprises
  counts <- vapply(
    seq_len(n),
    FUN = function(i) {
      d <- authors[i, names(row_names), drop = FALSE] |> as.matrix() |> t()
      rownames(d) <- row_names
      knitr::kable(d, format = "html", align = "r", col.names = NULL) |>
        kableExtra::kable_styling(
          bootstrap_options = c("row-border", "condensed"),
          full_width = FALSE,
          position = "left"
        ) |>
        as.character()
    },
    FUN.VALUE = character(1)
  )

  # get citations
  refs <- sprintf("%s (%s)", pubs$text_ref, pubs$year)
  names(refs) <- pubs$pub_id
  citations <- vapply(
    authors$pub_id,
    FUN = function(ids) {
      vapply(
        ids,
        FUN = function(id) {
          href <- sprintf("pub-%s.html", id)
          ref <- sprintf("<h4><a href='%s'>%s</a></h4>", href, refs[id])
          bib <- attr(unclass(pubs[id, "bibentry"])[[1]], which = "textVersion")
          paste(ref, bib, sep = "\n\n")
        },
        FUN.VALUE = character(1)
      ) |>
        paste(collapse = "\n\n")
    },
    FUN.VALUE = character(1)
  )

  # build scrollable cover strip for each author (omit pubs without a cover)
  cover_strips <- vapply(
    authors$pub_id,
    FUN = function(ids) {
      thumbs <- vapply(
        ids,
        FUN = function(id) {
          img_file <- file.path(coverdir, sprintf("pub-%s.jpg", id))
          if (!checkmate::test_file_exists(img_file, access = "r")) {
            return("")
          }
          sprintf(
            "<a href='pub-%s.html'><img src='pub-%s.jpg' style='height:300px; margin-right:4px;' alt='Cover' /></a>",
            id,
            id
          )
        },
        FUN.VALUE = character(1)
      )
      thumbs <- thumbs[nzchar(thumbs)]
      if (length(thumbs) == 0L) {
        return("")
      }
      paste0(
        "<div style='overflow-x:auto; white-space:nowrap; padding:4px 0; margin-bottom:20px;'>",
        paste(thumbs, collapse = "\n"),
        "</div>"
      )
    },
    FUN.VALUE = character(1)
  )

  # assemble each author's Rmd content
  contents <- vapply(
    seq_len(n),
    FUN = function(i) {
      plot_chunk <- sprintf(
        '```{r echo=FALSE, fig.height=3}\ninlpubs::plot_pubs(author_id = "%s")\n```',
        authors$author_id[i]
      )
      covers_section <- if (nzchar(cover_strips[i])) {
        paste("## Covers", "", cover_strips[i], "", sep = "\n")
      } else {
        ""
      }
      paste(
        headers[i],
        "",
        icons[i],
        "",
        "## Counts",
        "",
        counts[i],
        "",
        "## Publications",
        "",
        plot_chunk,
        "",
        covers_section,
        "",
        "## Citations",
        "",
        citations[i],
        "",
        sep = "\n"
      )
    },
    FUN.VALUE = character(1)
  )

  names <- sprintf("author-%s.Rmd", authors$author_id)
  replace_pages(contents, names, pattern = "^author-.*\\.Rmd$", destdir, quiet)

  invisible()
}


# Function to write publication webpages ----

make_pub_webpages <- function(pubs, destdir, coverdir, quiet) {
  # check arguments
  checkmate::assert_class(pubs, classes = c("pub", "data.frame"))
  checkmate::assert_path_for_output(destdir, overwrite = TRUE)
  checkmate::assert_string(coverdir)
  checkmate::assert_flag(quiet)

  n <- nrow(pubs)

  # get text references
  refs <- sprintf("%s (%s)", pubs$text_ref, pubs$year)
  names(refs) <- pubs$pub_id

  # make yaml headers
  headers <- vapply(
    seq_len(n),
    FUN = function(i) {
      # access the single bibentry with [[ ]] so $doi dispatches correctly
      doi_id <- pubs$bibentry[[i]]$doi

      # only build a DOI link when an identifier is present
      if (length(doi_id) > 0 && nzchar(doi_id)) {
        href <- paste0("https://doi.org/", doi_id)
        title <- "Launch the external DOI link, which leads to the publication content, in a separate browser tab."
        img <- "<img alt='DOI link' align='right' src='doi.svg' width='20' height='20' />"
        doi <- sprintf("<a href='%s' title='%s' target='_blank'>%s</a>", href, title, img)
      } else {
        doi <- ""
      }
      title <- sprintf("title: \"%s%s\"", refs[i], doi)
      paste("---", title, "---", sep = "\n")
    },
    FUN.VALUE = character(1)
  )

  # make cover images (only when the image file exists)
  covers <- vapply(
    pubs$pub_id,
    FUN = function(x) {
      file <- file.path(coverdir, sprintf("pub-%s.jpg", x))
      if (checkmate::test_file_exists(file, access = "r")) {
        sprintf(
          "<img class='cover-image' src='%s' alt='Cover image' align='right' style='width:300px;' />",
          basename(file)
        )
      } else {
        character(1)
      }
    },
    FUN.VALUE = character(1)
  )

  # get publication titles
  titles <- pubs$title

  # build linked author lists
  authors <- vapply(
    pubs$bibentry,
    FUN = function(x) {
      author <- format(x$author, include = c("given", "family"))
      author <- gsub(" ", "&nbsp;", author)
      ids <- names(x$author)
      for (i in seq_along(ids)) {
        href <- sprintf("author-%s.html", ids[i])
        author[i] <- sprintf("<a href='%s'>%s</a>", href, author[i])
      }
      n_auth <- length(author)
      if (n_auth > 2) {
        author[-n_auth] <- paste0(author[-n_auth], ",")
      }
      if (n_auth > 1) {
        author[n_auth] <- paste("and", author[n_auth])
      }
      paste(author, collapse = " ")
    },
    FUN.VALUE = character(1)
  )

  # get citations
  citations <- vapply(
    pubs$bibentry,
    FUN = function(x) {
      attr(unclass(x)[[1]], which = "textVersion")
    },
    FUN.VALUE = character(1)
  )

  # build collapsible BibTeX entries
  bibtexs <- vapply(
    pubs$bibentry,
    FUN = function(bib) {
      bibtex <- utils::toBibtex(bib) |> paste(collapse = "\n")
      template <- paste(
        "<a href='#bibtex' class='btn btn-default btn-xs' data-toggle='collapse' title='%s'>BibTeX</a>",
        "<div id='bibtex' class='collapse'><pre><code>%s</code></pre></div>",
        sep = "\n"
      )
      sprintf(template, "Switch between displaying and hiding the BibTeX citation.", bibtex)
    },
    FUN.VALUE = character(1)
  )

  # get abstracts and annotations
  abstracts <- pubs$abstract |> stringi::stri_unescape_unicode()
  annotations <- pubs$annotation |> stringi::stri_unescape_unicode()

  # build annotation source links
  txt <- refs[pubs$annotation_src]
  hrefs <- sprintf("pub-%s.html", names(txt))
  annotation_srcs <- sprintf("<a href='%s'>%s</a>", hrefs, txt)
  annotation_srcs[is.na(txt)] <- NA_character_

  # assemble each publication's Rmd content
  contents <- vapply(
    seq_len(n),
    FUN = function(i) {
      parts <- c(
        headers[i],
        "",
        paste("## Title", covers[i]),
        "",
        titles[i],
        "",
        "## Authors",
        "",
        authors[i],
        "",
        "## Citation",
        "",
        citations[i],
        "",
        bibtexs[i],
        ""
      )
      if (!is.na(abstracts[i])) {
        parts <- c(parts, "## Abstract", "", abstracts[i], "")
      }
      if (!is.na(annotations[i])) {
        parts <- c(
          parts,
          "## Annotation",
          "",
          annotations[i],
          "",
          paste("--- From", annotation_srcs[i]),
          ""
        )
      }
      paste(parts, collapse = "\n")
    },
    FUN.VALUE = character(1)
  )

  names <- sprintf("pub-%s.Rmd", pubs$pub_id)
  replace_pages(contents, names, pattern = "^pub-.*\\.Rmd$", destdir, quiet)

  invisible()
}

#' Search a Term-Frequency Table
#'
#' @description Search a term-frequency table by pattern or frequency.
#'
#' @param x 'character' string.
#'   Regular-expression pattern to search for among the table's terms.
#'
#' @param data 'term' and 'data.frame' class.
#'   Term-frequency table to search, with columns `term`, `pub_id`, and `freq`.
#'   Defaults to the term frequencies from the INLPO publications; see the [`terms`]
#'   dataset for details.
#'
#' @param ignore_case 'logical' flag.
#'   Whether to ignore character case when matching. Defaults to `TRUE`.
#'
#' @param ...
#'   Additional arguments passed to [grep()] (for example, `fixed` or `perl`).
#'
#' @param low_freq 'numeric' number.
#'   Lower frequency bound; terms occurring fewer than this many times are excluded.
#'   Defaults to 1.
#'
#' @param high_freq 'numeric' number.
#'   Upper frequency bound; terms occurring more than this many times are excluded.
#'   Defaults to `Inf` (no upper limit).
#'
#' @param simplify 'logical' flag.
#'   Whether to return only the unique publication identifiers (`TRUE`, the default)
#'   rather than the full table of matching rows.
#'
#' @return If `simplify = TRUE`, a 'character' vector of unique publication
#'   identifiers, ordered by the most frequent matching term first. If
#'   `simplify = FALSE`, a 'data.frame' of matching rows sorted by decreasing
#'   frequency.
#'
#' @author J.C. Fisher, U.S. Geological Survey, Idaho Water Science Center
#'
#' @export
#'
#' @seealso [`mine_text`] to build a term-frequency table.
#'
#' @examples
#' search_terms("mlms")
#'
#' out <- search_terms("mlms", simplify = FALSE)
#' head(out)

search_terms <- function(
  x,
  data = inlpubs::terms,
  ignore_case = TRUE,
  ...,
  low_freq = 1,
  high_freq = Inf,
  simplify = TRUE
) {
  # check arguments
  checkmate::assert_string(x, na.ok = FALSE)
  checkmate::assert_class(data, classes = c("term", "data.frame"))
  checkmate::assert_flag(ignore_case)
  checkmate::assert_number(low_freq, lower = 1)
  checkmate::assert_number(high_freq, lower = low_freq)
  checkmate::assert_flag(simplify)

  # select rows whose term matches the search pattern
  matched <- grepl(pattern = x, data$term, ignore.case = ignore_case, ...)
  out <- data[matched, ]

  # restrict to the requested frequency range
  in_range <- out$freq >= low_freq & out$freq <= high_freq
  out <- out[in_range, ]

  # sort by decreasing frequency, then publication identifier
  idxs <- order(out$freq, out$pub_id, decreasing = c(TRUE, FALSE), method = "radix")
  out <- out[idxs, ]

  # drop unused factor levels
  out <- droplevels(out)

  # optionally reduce to unique publication identifiers
  if (simplify) {
    out <- out$pub_id |>
      as.character() |>
      unique()
  }

  out
}

# setup tinytest for checkmate functionality
library("tinytest")
library("checkmate")
using("checkmate")

# ------ minimal term dataset fixture ------

d <- structure(
  data.frame(
    term = factor(
      c("water", "water", "aquifer", "aquifer", "nitrate")
    ),
    pub_id = factor(
      c("PubA", "PubB", "PubA", "PubC", "PubB")
    ),
    freq = c(10L, 3L, 5L, 8L, 2L)
  ),
  class = c("term", "data.frame")
)

# ------ simplify = TRUE (default) ------

# returns a character vector
out <- search_terms("water", data = d)
expect_true(is.character(out))

# returns only unique pub_ids
expect_equal(length(out), length(unique(out)))

# correct pub_ids matched
expect_true(all(c("PubA", "PubB") %in% out))
expect_false("PubC" %in% out)

# ordered by most frequent matching term first (PubA freq=10 before PubB freq=3)
expect_equal(out[1], "PubA")

# ------ simplify = FALSE ------

out_df <- search_terms("water", data = d, simplify = FALSE)
expect_true(is.data.frame(out_df))
expect_equal(colnames(out_df), c("term", "pub_id", "freq"))
expect_true(nrow(out_df) > 0L)

# sorted by decreasing frequency
expect_true(all(diff(out_df$freq) <= 0L))

# ------ ignore_case ------

out_upper <- search_terms("WATER", data = d, ignore_case = TRUE)
out_exact <- search_terms("WATER", data = d, ignore_case = FALSE)
expect_true(length(out_upper) > 0L)
expect_equal(length(out_exact), 0L)

# ------ frequency filtering ------

# low_freq excludes terms below threshold
out_low <- search_terms("water", data = d, low_freq = 5)
expect_false("PubB" %in% out_low) # PubB has freq=3, below threshold
expect_true("PubA" %in% out_low) # PubA has freq=10

# high_freq excludes terms above threshold
out_high <- search_terms("water", data = d, high_freq = 5)
expect_false("PubA" %in% out_high) # PubA has freq=10, above threshold
expect_true("PubB" %in% out_high) # PubB has freq=3

# ------ no matches returns empty result ------

expect_equal(length(search_terms("zzznomatch", data = d)), 0L)
expect_equal(nrow(search_terms("zzznomatch", data = d, simplify = FALSE)), 0L)

# ------ regex pattern ------

out_regex <- search_terms("wat|nit", data = d)
expect_true("PubA" %in% out_regex)
expect_true("PubB" %in% out_regex)

# ------ invalid arguments ------

expect_error(search_terms(NA_character_, data = d))
expect_error(search_terms("water", data = d, low_freq = 0))
expect_error(search_terms("water", data = d, low_freq = 5, high_freq = 3))
expect_error(search_terms("water", data = data.frame()))

#' Contributing Authors to INLPO Publications
#'
#' @description Authors who have contributed to publications by the U.S. Geological
#'   Survey (USGS), Idaho Water Science Center, Idaho National Laboratory Project
#'   Office (INLPO).
#'
#' @format An object of class 'author' that inherits from the 'data.frame' class and
#'   contains 267 rows and the following 10 columns:
#'   \describe{
#'     \item{`author_id`}{Unique identifier for the author, composed of the author's
#'       first initial and last name in lowercase (for example, `"jfisher"`).}
#'     \item{`name`}{Author name, surname first, followed by initials or given name.}
#'     \item{`person`}{A named list of [person][utils::person] objects holding
#'       contact details such as email address and
#'       [ORCID](https://orcid.org/) identifier. Names match `author_id`.}
#'     \item{`pub_id`}{A named list of character vectors giving the identifier(s) of
#'       the publication(s) the author contributed to; each value is a foreign key
#'       referencing the `pub_id` primary key of the [`pubs`] table.}
#'     \item{`total_pub`}{Total number of publications.}
#'     \item{`single_authored`}{Number of single-authored publications.}
#'     \item{`multi_authored`}{Number of multi-authored publications.}
#'     \item{`first_authored`}{Number of multi-authored publications in which the
#'       author appears as the first author.}
#'     \item{`first_year`}{First year in which the author published.}
#'     \item{`last_year`}{Most recent year in which the author published.}
#'   }
#'
#' @source Curated by INLPO staff.
#'
#' @seealso [`pubs`] for the publications referenced by `pub_id`.
#'
#' @keywords datasets
#'
#' @examples
#' # Subset Jason Fisher's information and display its structure:
#' author <- authors["jfisher", ]
#' str(author, max.level = 3, width = 75, strict.width = "cut")
#'
#' # Print the author's given name:
#' author$person |> format(include = "given")

"authors"

#' Bibliographic Records for INLPO Publications
#'
#' @description Bibliographic information for reports, articles, maps, and theses
#'   related to scientific monitoring and research conducted by the U.S. Geological
#'   Survey (USGS), Idaho Water Science Center, Idaho National Laboratory Project
#'   Office (INLPO).
#'
#' @format An object of class 'pub' that inherits from the 'data.frame' class and
#'   includes the following columns:
#'   \describe{
#'     \item{`pub_id`}{Unique identifier for the publication (the table's primary
#'       key), in upper camel case (for example, `"FisherOthers2012"`).}
#'     \item{`institution`}{Name of the institution that published and/or sponsored
#'       the report.}
#'     \item{`type`}{Type of publication, such as `"Open-File Report"`,
#'       `"Data Series"`, or `"Manual"`.}
#'     \item{`text_ref`}{Text reference (the in-text citation) excluding the year of
#'       publication, such as `"Dietz"` or `"Treinen and others"`.}
#'     \item{`year`}{Year of publication.}
#'     \item{`author_id`}{A named list of character vectors giving the identifier(s)
#'       of the contributing author(s), in author order; each value is a foreign key
#'       referencing the `author_id` primary key of the [`authors`] table. List names
#'       match `pub_id`.}
#'     \item{`title`}{Title of publication.}
#'     \item{`bibentry`}{A named list of [bibentry][utils::bibentry] objects holding
#'       the full bibliographic record, from which formatted citations and BibTeX
#'       entries are derived. List names match `pub_id`.}
#'     \item{`abstract`}{Abstract of publication, or `NA` when unavailable.}
#'     \item{`annotation`}{Annotation summarizing the publication, or `NA` when none
#'       is provided.}
#'     \item{`annotation_src`}{Identifier (`pub_id`) of the publication the annotation
#'       was taken from, or `NA` when not applicable
#'       (Knobel and others, 2005; Bartholomay, 2022).}
#'     \item{`files`}{A named list of character vectors naming the file(s) associated
#'       with the publication, primary document first, or `NULL` when no files are
#'       recorded. List names match `pub_id`.}
#'   }
#'
#' @source Many of these publications are available through the
#'   [USGS Publications Warehouse](https://pubs.usgs.gov/).
#'
#' @references Bartholomay, R.C., 2022, Historical development of the U.S. Geological Survey
#'   hydrological monitoring and investigative programs at the Idaho National Laboratory, Idaho, 2002-2020:
#'   U.S. Geological Survey Open-File Report 2022-1027 (DOE/ID-22256), 54 p., \doi{10.3133/ofr20221027}.
#' @references Knobel, L.L., Bartholomay, R.C., and Rousseau, J.P., 2005,
#'   Historical development of the U.S. Geological Survey hydrologic monitoring and investigative programs
#'   at the Idaho National Engineering and Environmental Laboratory, Idaho, 1949 to 2001:
#'   U.S. Geological Survey Open-File Report 2005--1223 (DOE/ID--22195), 93 p.,
#'   \doi{10.3133/ofr20051223}.
#'
#' @seealso [`authors`] for the contributing authors referenced by `author_id`,
#'   [`terms`] for term frequencies derived from publication text, and
#'   [`search_terms`] to query publications by term.
#'
#' @keywords datasets
#'
#' @examples
#' # Subset Fisher and others (2012) and display its structure:
#' id <- "FisherOthers2012"
#' pub <- pubs[id, ]
#' str(pub, max.level = 3, width = 75, strict.width = "cut")
#'
#' # Print the suggested citation:
#' attr(unclass(pub$bibentry[[1]])[[1]], which = "textVersion")
#'
#' # Print the authors' full names:
#' format(pub$bibentry[[1]]$author, include = c("given", "family"))
#'
#' # Print the abstract:
#' pub$abstract

"pubs"

#' Build a Term-Frequency Table
#'
#' @description Build a term-frequency table from document text.
#'   A term is a single word or, when n-gram tokenization is enabled, an ordered
#'   sequence of words.
#'
#' @param docs 'list' or 'character' vector.
#'   Document text to analyze. Each element holds the extracted text of a single
#'   document and may itself be a character vector (its pieces are concatenated).
#'   `NA` values are dropped. If unnamed, elements are numbered sequentially.
#'
#' @param ngmin,ngmax 'integer' numbers.
#'   Minimum and maximum number of words per term (n-gram). The default
#'   (`ngmin = ngmax = 1`) counts single words. Values greater than one require the
#'   \pkg{RWeka} package and that the `JAVA_HOME` environment variable point to a
#'   Java installation. Recommended for a single document only.
#'
#' @param sparse 'numeric' number greater than 0 and less than 1, or `NULL`.
#'   Maximum allowed sparsity: terms appearing in a smaller proportion of documents
#'   are removed. For example, `sparse = 0.99` removes terms more sparse than 0.99,
#'   while `sparse = 0.01` retains only terms appearing in nearly every document.
#'   The default (`NULL`) disables sparse-term removal.
#'
#' @details Documents are cleaned before counting: URLs, escaped unicode sequences,
#'   HTML tags, and selected punctuation are removed; text is lowercased; numbers,
#'   English and SMART stopwords, and remaining punctuation are stripped; and
#'   whitespace is collapsed. HTML entities are decoded when the \pkg{textutils}
#'   package is available. When n-grams are requested, terms containing a repeated
#'   word are discarded.
#'
#' @return A 'data.frame' term-frequency table with columns `term`, `pub_id`, and
#'   `freq`, giving the number of times each term occurs in each document. Rows with
#'   a zero count are omitted, and the table is sorted by term (ascending) then
#'   frequency (descending).
#'
#' @author J.C. Fisher, U.S. Geological Survey, Idaho Water Science Center
#'
#' @seealso [`search_terms`] to search within the resulting table, and
#'   [`make_wordcloud`] to create a word cloud from it.
#'
#' @export
#'
#' @examples
#' d <- c(
#'   "The quick brown fox jumps over the lazy lazy dog.",
#'   "Pack my brown box.",
#'   "Jazz fly brown dog."
#' ) |>
#'   mine_text()
#'
#' d <- list(
#'   "A" = "The quick brown fox jumps over the lazy lazy dog.",
#'   "B" = c("Pack my brown box.", NA, "Jazz fly brown dog."),
#'   "C" = NA_character_
#' ) |>
#'   mine_text()

mine_text <- function(
  docs,
  ngmin = 1,
  ngmax = ngmin,
  sparse = NULL
) {
  # check arguments
  checkmate::assert(
    checkmate::check_list(docs),
    checkmate::check_character(docs)
  )
  checkmate::assert_count(ngmin, positive = TRUE)
  checkmate::assert_int(ngmax, lower = ngmin)
  checkmate::assert_number(sparse, lower = 0.001, upper = 0.999, null.ok = TRUE)

  # check system dependencies for n-gram tokenization up front
  if (ngmax > 1L) {
    if (!checkmate::test_directory_exists(Sys.getenv("JAVA_HOME"), access = "r")) {
      stop("Word n-grams require JAVA_HOME to point to a Java installation", call. = FALSE)
    }
    if (!requireNamespace("RWeka", quietly = TRUE)) {
      stop("Word n-grams require the 'RWeka' package", call. = FALSE)
    }
  }

  # assign names to unnamed documents
  if (is.null(names(docs))) {
    names(docs) <- as.character(seq_along(docs))
  }

  # concatenate each document's pieces into a single string, dropping NA
  docs <- vapply(
    docs,
    FUN = function(x) {
      stats::na.omit(x) |>
        as.character() |>
        paste(collapse = " ")
    },
    FUN.VALUE = character(1)
  )

  # decode HTML entities when textutils is available
  if (requireNamespace("textutils", quietly = TRUE)) {
    docs <- textutils::HTMLdecode(docs)
  }

  # define transformation functions
  remove_url <- tm::content_transformer(
    function(x) gsub("(f|ht)tp(s?)://\\S+", "", x, perl = TRUE)
  )
  remove_pat <- tm::content_transformer(
    function(x, pattern) gsub(pattern, " ", x)
  )

  # build and clean the volatile corpus
  corpus <- tm::VectorSource(docs) |>
    tm::VCorpus() |>
    tm::tm_map(remove_url) |>
    tm::tm_map(remove_pat, "\\\\u[0-9A-Fa-f]{4}") |>
    tm::tm_map(remove_pat, "\\\\n") |>
    tm::tm_map(remove_pat, "<.*?>") |>
    tm::tm_map(remove_pat, "/") |>
    tm::tm_map(remove_pat, "@") |>
    tm::tm_map(remove_pat, "\\|") |>
    tm::tm_map(tm::content_transformer(tolower)) |>
    tm::tm_map(tm::removeNumbers) |>
    tm::tm_map(tm::removeWords, tm::stopwords("english")) |>
    tm::tm_map(tm::removeWords, tm::stopwords("SMART")) |>
    tm::tm_map(tm::removePunctuation, preserve_intra_word_dashes = TRUE) |>
    tm::tm_map(tm::stripWhitespace)

  # tag each corpus document with its identifier
  for (i in seq_along(corpus)) {
    corpus[[i]]$meta$id <- names(docs)[i]
  }

  # define the n-gram tokenizer when more than single words are requested
  control <- list()
  if (ngmax > 1L) {
    control <- list(
      "tokenize" = function(x) {
        RWeka::NGramTokenizer(
          x,
          control = RWeka::Weka_control(min = ngmin, max = ngmax)
        )
      }
    )
  }

  # coerce the corpus to a term-document matrix
  tdm <- tm::TermDocumentMatrix(corpus, control = control)

  # remove sparse terms
  if (!is.null(sparse)) {
    tdm <- tm::removeSparseTerms(tdm, sparse = sparse)
  }

  # convert to a long-format frequency table
  d <- as.table(tdm) |>
    as.data.frame(stringsAsFactors = TRUE)
  colnames(d) <- c("term", "pub_id", "freq")
  d$freq <- as.integer(d$freq)

  # drop zero-frequency rows
  d <- d[d$freq > 0, ]

  # return early when no terms remain (e.g. all-NA input)
  if (nrow(d) == 0L) {
    return(droplevels(d))
  }

  # for n-grams, drop terms containing a repeated word
  if (ngmax > 1L) {
    has_dup <- as.character(d$term) |>
      strsplit(split = " ") |>
      vapply(FUN = anyDuplicated, FUN.VALUE = integer(1)) |>
      as.logical()
    d <- d[!has_dup, ]
  }

  # sort by term (ascending), then frequency (descending)
  idxs <- order(d$term, d$freq, decreasing = c(FALSE, TRUE), method = "radix")
  d <- d[idxs, ]

  # clear row names and drop unused factor levels
  rownames(d) <- NULL
  droplevels(d)
}

# setup tinytest for checkmate functionality
library("tinytest")
library("checkmate")
using("checkmate")

# test that 3-grams result in 3 word tokens, requires JAVA system dependency
if (test_directory_exists(Sys.getenv("JAVA_HOME"), access = "r")) {
  d <- mine_text(docs = pubs$title, ngmin = 3L)
  nwords <- as.character(d$term) |>
    strsplit(" ") |>
    vapply(FUN = length, FUN.VALUE = integer(1))
  all(nwords == 3L) |> expect_true()
}

# setup tinytest for checkmate functionality
library("tinytest")
library("checkmate")
using("checkmate")

# ------ fixtures ------

docs_named <- list(
  "A" = "The quick brown fox jumps over the lazy lazy dog.",
  "B" = c("Pack my brown box.", NA, "Jazz fly brown dog."),
  "C" = NA_character_
)

docs_unnamed <- c(
  "The quick brown fox jumps over the lazy lazy dog.",
  "Pack my brown box.",
  "Jazz fly brown dog."
)

d_named <- mine_text(docs_named)
d_unnamed <- mine_text(docs_unnamed)

# ------ return structure ------

# result is a data.frame
expect_true(is.data.frame(d_named))

# has exactly the three expected columns
expect_equal(colnames(d_named), c("term", "pub_id", "freq"))

# term and pub_id are factors, freq is integer
expect_true(is.factor(d_named$term))
expect_true(is.factor(d_named$pub_id))
expect_true(is.integer(d_named$freq))

# no zero-frequency rows
expect_true(all(d_named$freq > 0L))

# no NA values in any column
expect_false(anyNA(d_named))

# ------ pub_id assignment ------

# named docs: pub_ids match the list names (excluding all-NA doc "C")
expect_true(all(levels(d_named$pub_id) %in% c("A", "B", "C")))
expect_false("C" %in% as.character(d_named$pub_id)) # all-NA produces no rows

# unnamed docs: pub_ids are sequential integers as character
expect_true(all(levels(d_unnamed$pub_id) %in% c("1", "2", "3")))

# ------ term counts ------

# "lazy" appears twice in doc "A", so its freq should be 2
lazy_a <- d_named[d_named$term == "lazy" & d_named$pub_id == "A", "freq"]
expect_equal(lazy_a, 2L)

# "brown" appears in docs "A" and "B"
brown_pubs <- as.character(d_named$pub_id[d_named$term == "brown"])
expect_true("A" %in% brown_pubs)
expect_true("B" %in% brown_pubs)

# ------ stopword removal ------

# common English stopwords must not appear as terms
sw <- c("the", "my", "over")
expect_false(any(sw %in% as.character(d_named$term)))

# ------ number removal ------

d_nums <- mine_text(list("x" = "There are 3 cats and 42 dogs."))
expect_false(any(grepl("[0-9]", as.character(d_nums$term))))

# ------ URL removal ------

d_url <- mine_text(list("x" = "Visit https://example.com for details water quality."))
expect_false(any(grepl("http", as.character(d_url$term))))

# ------ sort order ------

# terms must be in ascending alphabetical order within each pub_id
terms_a <- as.character(d_named$term[d_named$pub_id == "A"])
expect_equal(terms_a, sort(terms_a))

# for a term appearing in multiple docs, rows with the same term must be
# sorted by descending frequency
if (nrow(d_named) > 1) {
  for (trm in unique(as.character(d_named$term))) {
    freqs <- d_named$freq[d_named$term == trm]
    expect_true(all(diff(freqs) <= 0L))
  }
}

# ------ all-NA document produces no rows ------

d_allna <- mine_text(list("only" = NA_character_))
expect_equal(nrow(d_allna), 0L)

# ------ sparse argument ------

# with sparse = 0.5, terms appearing in fewer than half the docs are removed;
# "brown" appears in 2 of 3 docs so it should survive
d_sparse <- mine_text(docs_unnamed, sparse = 0.5)
expect_true("brown" %in% as.character(d_sparse$term))

# with a very low sparse threshold (term must appear in nearly every doc),
# a term that appears in only one doc should be removed
d_strict <- mine_text(docs_unnamed, sparse = 0.01)
# "box" only appears in doc 2, so it must not survive
expect_false("box" %in% as.character(d_strict$term))

# ------ invalid argument errors ------

expect_error(mine_text(123))
expect_error(mine_text("text", ngmin = 0L))
expect_error(mine_text("text", ngmin = 2L, ngmax = 1L))
expect_error(mine_text("text", sparse = 0))
expect_error(mine_text("text", sparse = 1))

#' Obtain Text from a PDF Document
#'
#' @description Extract text from a PDF document and write it to a plain-text
#'   file, one page per line. Pages with an extractable text layer are read
#'   directly; pages without one (e.g. scanned images) fall back to optical
#'   character recognition (OCR). The text is normalized to ASCII before writing.
#'
#'   Requires the \pkg{pdftools}, \pkg{tesseract}, and \pkg{callr} packages;
#'   image preprocessing additionally requires \pkg{magick}.
#'
#' @details
#'   Direct extraction is attempted first, as it is faster and more accurate than
#'   OCR for born-digital PDFs; a page is OCR'd only when its text layer yields
#'   fewer than `min_chars` non-whitespace characters, or when `force_ocr` is
#'   `TRUE`. Poppler reads and page rendering run in isolated child processes, so
#'   a renderer crash surfaces as a normal R error rather than aborting the
#'   session.
#'
#'   Directly extracted text is reordered into human reading order from word
#'   bounding boxes: full-width headers and captions are set aside, then the body
#'   is split into columns only on strong evidence of a multi-column layout. OCR
#'   uses Tesseract's LSTM engine, run in parallel across `cores` workers when
#'   several pages need it. The result is normalized to strictly ASCII: selected
#'   scientific symbols become ASCII tokens, accents are transliterated, line-break
#'   hyphenation is rejoined, and quotes, control characters, and repeated
#'   punctuation and whitespace are cleaned up.
#'
#' @param input 'character' string.
#'   File path to the PDF document.
#'
#' @param output 'character' string.
#'   File path for the text file to write. If the path lacks a `.txt` or `.text`
#'   extension, `.txt` is appended. Defaults to a temporary file.
#'
#' @param dpi 'integer' number between 100 and 1200.
#'   Resolution, in dots per inch, at which pages are rendered before OCR.
#'   Higher values improve OCR accuracy at the cost of speed; 300 (the default)
#'   suits most documents, while 400 to 600 may help with small fonts or poor
#'   scans. Ignored for pages read directly from the text layer.
#'
#' @param psm 'integer' number between 0 and 13.
#'   Tesseract Page Segmentation Mode, describing the expected page layout.
#'   The default of 1 segments the page automatically and detects orientation and
#'   script, which suits multi-column documents.
#'
#' @param force_ocr 'logical' flag.
#'   Whether to OCR every page regardless of its text layer. Use this when the
#'   embedded text is known to be unreliable (e.g. fonts that extract as gibberish,
#'   which `min_chars` cannot catch because the gibberish is long enough to pass).
#'   Defaults to `FALSE`.
#'
#' @param preprocess 'logical' flag.
#'   Whether to grayscale, deskew, despeckle, and normalize each page image before
#'   OCR, which improves accuracy on scanned documents at a modest per-page cost.
#'   Requires the \pkg{magick} package. Applies only to pages that are OCR'd.
#'   Defaults to `FALSE`.
#'
#' @param min_chars 'integer' number.
#'   Minimum number of non-whitespace characters a page's extracted text must
#'   have to be accepted without OCR. Pages below this threshold are sent to OCR.
#'   Defaults to 20.
#'
#' @param cores 'integer' number.
#'   Number of parallel workers to use for OCR. When `NULL` (the default), the
#'   count is taken from the `OCR_CORES` environment variable if set (useful for
#'   honoring a container CPU quota), otherwise from one fewer than the number of
#'   detected cores.
#'
#' @param timeout 'numeric' number.
#'   Maximum time, in seconds, to allow each child-process call to run before it
#'   is aborted. Applied independently to the text-extraction call and the
#'   page-rendering call. Defaults to 120.
#'
#' @return Returns the path to the text file, invisibly.
#'   Each page of the PDF is written as a separate line.
#'
#' @author J.C. Fisher, U.S. Geological Survey, Idaho Water Science Center
#'
#' @export
#'
#' @seealso [`add_content`] to add extracted text to the \pkg{inlpubs}-package corpus.
#'
#' @examples
#' \dontrun{
#'   input <- system.file("extdata", "test.pdf", package = "inlpubs")
#'   path <- get_pdf_text(input)
#'
#'   unlink(path)
#' }

get_pdf_text <- function(
  input,
  output = tempfile(fileext = ".txt"),
  dpi = 300,
  psm = 1,
  force_ocr = FALSE,
  preprocess = FALSE,
  min_chars = 20,
  cores = NULL,
  timeout = 120
) {
  # check system dependencies
  if (!requireNamespace("pdftools", quietly = TRUE)) {
    stop("Reading a PDF requires the 'pdftools' package", call. = FALSE)
  }
  if (!requireNamespace("tesseract", quietly = TRUE)) {
    stop("OCR requires the 'tesseract' package", call. = FALSE)
  }
  if (!requireNamespace("callr", quietly = TRUE)) {
    stop("Crash-safe PDF reading requires the 'callr' package", call. = FALSE)
  }
  if (preprocess && !requireNamespace("magick", quietly = TRUE)) {
    stop("Image preprocessing requires the 'magick' package", call. = FALSE)
  }

  # resolve default core count, honoring an env override (e.g. container quota)
  # and guarding against detectCores() returning NA
  if (is.null(cores)) {
    env <- Sys.getenv("OCR_CORES", unset = NA)
    if (!is.na(env)) {
      cores <- suppressWarnings(as.integer(env))
      if (is.na(cores)) {
        stop(
          "Environment variable 'OCR_CORES' must be a positive integer, got: ",
          sQuote(env),
          call. = FALSE
        )
      }
    } else {
      n <- parallel::detectCores()
      cores <- if (is.na(n)) 1L else max(1L, n - 1L)
    }
  }

  # check arguments
  input <- path.expand(input) |> normalizePath(winslash = "/", mustWork = FALSE)
  checkmate::assert_file_exists(input, access = "r", extension = "pdf")
  output <- path.expand(output) |> normalizePath(winslash = "/", mustWork = FALSE)
  checkmate::assert_path_for_output(output, overwrite = TRUE)
  checkmate::assert_int(dpi, lower = 100, upper = 1200)
  checkmate::assert_int(psm, lower = 0, upper = 13)
  checkmate::assert_flag(force_ocr)
  checkmate::assert_flag(preprocess)
  checkmate::assert_int(min_chars, lower = 0)
  checkmate::assert_int(cores, lower = 1)
  checkmate::assert_number(timeout, lower = 1)

  # ensure the output has a text-file extension
  if (!grepl("\\.(txt|text)$", tolower(output))) {
    output <- paste0(output, ".txt")
  }

  # Poppler can hard-crash the R process on certain files (e.g. STATUS_STACK_
  # BUFFER_OVERRUN on Windows) below the R level, where tryCatch cannot catch it.
  # Running the reads in a child process turns such a crash into a normal R error
  # instead of taking down the session or aborting a batch run. pdf_data() gives
  # per-word boxes used to reconstruct reading order.
  read <- tryCatch(
    callr::r(
      func = function(pdf, get_text) {
        info <- pdftools::pdf_info(pdf)
        list(
          pages = info$pages,
          data = if (get_text) pdftools::pdf_data(pdf) else NULL
        )
      },
      args = list(pdf = input, get_text = !force_ocr),
      timeout = timeout
    ),
    error = function(e) NULL
  )
  if (is.null(read)) {
    stop(
      "Could not read PDF (the renderer crashed or timed out on this file): ",
      sQuote(basename(input)),
      call. = FALSE
    )
  }

  # initialize one text slot per page so OCR results can be slotted back by index
  n_pages <- read$pages
  text <- character(n_pages)

  # try direct text-layer extraction first; cheap and accurate for born-digital
  # PDFs. force_ocr skips this and flags every page for OCR.
  if (force_ocr) {
    needs_ocr <- rep(TRUE, n_pages)
  } else {
    # reconstruct reading order per page from the word boxes
    direct <- vapply(read$data, .reconstruct_page, character(1))
    text <- direct
    # a page needs OCR if its text layer is empty or implausibly short
    needs_ocr <- nchar(trimws(direct)) < min_chars
  }

  # OCR only the pages that need it
  if (any(needs_ocr)) {
    text[which(needs_ocr)] <- .run_ocr_pages(
      input = input,
      pages = which(needs_ocr),
      dpi = dpi,
      psm = psm,
      preprocess = preprocess,
      cores = cores,
      timeout = timeout
    )
  }

  # normalize the text only when there is something to convert; guard against an
  # empty document so encoding handling does not error on a zero-length input
  if (length(text) > 0 && any(nzchar(text))) {
    # preserve high-value scientific symbols before the blanket ASCII squash
    # would drop them to spaces
    text <- gsub("\u00b5", "u", text) # micro sign -> u (e.g. ug/L)
    text <- gsub("\u03bc", "u", text) # Greek small mu -> u
    text <- gsub("\u00b0", " deg ", text) # degree sign
    text <- gsub("\u00b1", "+/-", text) # plus-minus sign

    # transliterate accents (keeps names searchable, e.g. "Mendez"), then force
    # the remainder to ASCII
    text <- stringi::stri_trans_general(text, "Latin-ASCII")
    text <- iconv(text, from = "UTF-8", to = "ASCII", sub = " ")

    # rejoin hyphenation BEFORE control chars are flattened, so the newline is
    # still present. The hyphen must end a line (directly precede \n) to count,
    # which rejoins true splits ("exam-\nple", "well-\n7B") but leaves mid-line
    # "890- square" alone; an uppercase continuation is treated as a real compound.
    text <- gsub("([a-z0-9])-\n[[:blank:]]*([a-z0-9])", "\\1\\2", text)
    text <- gsub("[[:cntrl:]]", " ", text)
    text <- gsub("[\u2018\u2019]", "'", text) # curly -> straight single quotes
    text <- gsub("[\u201C\u201D]", "\"", text) # curly -> straight double quotes
    text <- gsub("\\.{2,}", ".", text) # collapse repeated periods
    text <- gsub("\\s+", " ", text)
    text <- trimws(text)
  }

  # write text, one page per line. useBytes keeps the already-ASCII content from
  # being re-encoded under a non-UTF-8 native locale.
  writeLines(text, con = output, useBytes = TRUE)

  # confirm the file was written
  checkmate::assert_file_exists(output, access = "r")

  invisible(output)
}

# Reconstruct a page's reading order from word boxes returned by pdf_data().
# Full-width rows (headers, captions, table rows) are emitted in place and
# excluded from gutter detection. Splitting requires strong evidence (wide
# gutter, both sides substantial); otherwise the page reads single-column.
.reconstruct_page <- function(words) {
  if (is.null(words) || nrow(words) == 0L) {
    return("") # blank or image-only page
  }

  bucket <- 20
  min_gutter_buckets <- 3L # gutter = this many consecutive empty x-bands
  min_words_per_col <- 5L
  line_tol <- 3 # words within this y-distance (pts) share a line
  full_frac <- 0.6 # row spanning >= this fraction of extent = full-width

  # horizontal extent from the word boxes
  x_min <- min(words$x)
  x_max <- max(words$x + words$width)
  span <- x_max - x_min
  mid <- (x_min + x_max) / 2

  # group words into rows by y, then flag rows that cross the midline and span
  # most of the extent as full-width
  ord <- order(words$y, words$x)
  w <- words[ord, , drop = FALSE]
  row_id <- cumsum(c(TRUE, abs(diff(w$y)) > line_tol))
  rows <- split(seq_len(nrow(w)), row_id)
  is_full <- vapply(
    rows,
    function(idx) {
      rx_min <- min(w$x[idx])
      rx_max <- max(w$x[idx] + w$width[idx])
      (rx_min < mid) && (rx_max > mid) && ((rx_max - rx_min) >= span * full_frac)
    },
    logical(1)
  )

  # detect the gutter from body words only (full-width rows removed)
  center <- .detect_gutter(w, rows, is_full, x_min, span, bucket, min_gutter_buckets)

  # lines from a word set: top to bottom, left to right within a line
  emit_lines <- function(idx) {
    if (length(idx) == 0L) {
      return(character(0))
    }
    sub <- w[idx, , drop = FALSE]
    sub <- sub[order(sub$y, sub$x), , drop = FALSE]
    lid <- cumsum(c(TRUE, abs(diff(sub$y)) > line_tol))
    out <- character(0)
    for (id in unique(lid)) {
      ln <- sub[lid == id, , drop = FALSE]
      ln <- ln[order(ln$x), , drop = FALSE]
      out <- c(out, paste(ln$text, collapse = " "))
    }
    out
  }

  # a body run: column-split when a gutter was found and both sides substantial
  emit_body <- function(idx) {
    if (length(idx) == 0L) {
      return(character(0))
    }
    if (!is.na(center)) {
      left <- idx[w$x[idx] < center]
      right <- idx[w$x[idx] >= center]
      if (length(left) >= min_words_per_col && length(right) >= min_words_per_col) {
        return(c(emit_lines(left), emit_lines(right)))
      }
    }
    emit_lines(idx)
  }

  # walk top to bottom: full-width rows in place, body runs column-split between
  lines <- character(0)
  body_buffer <- integer(0)
  flush_body <- function() {
    if (length(body_buffer) > 0L) {
      lines <<- c(lines, emit_body(body_buffer))
      body_buffer <<- integer(0)
    }
  }
  for (k in seq_along(rows)) {
    idx <- rows[[k]]
    if (is_full[k]) {
      flush_body()
      lines <- c(lines, emit_lines(idx))
    } else {
      body_buffer <- c(body_buffer, idx)
    }
  }
  flush_body()

  paste(lines, collapse = "\n")
}

# Find the x-center of the column gutter from body (non-full-width) word boxes.
# Returns NA_real_ when no strong gutter is found.
.detect_gutter <- function(w, rows, is_full, x_min, span, bucket, min_gutter_buckets) {
  body_idx <- unlist(rows[!is_full], use.names = FALSE)
  if (length(body_idx) == 0L) {
    return(NA_real_)
  }

  b0 <- floor(w$x[body_idx] / bucket)
  b1 <- floor((w$x[body_idx] + w$width[body_idx]) / bucket)
  occupied <- integer(0)
  for (i in seq_along(b0)) {
    occupied <- c(occupied, b0[i]:b1[i])
  }
  occupied <- unique(occupied)

  lo <- floor((x_min + span * 0.33) / bucket)
  hi <- floor((x_min + span * 0.66) / bucket)
  best_run <- 0L
  run <- 0L
  run_start <- NA_integer_
  center <- NA_real_
  for (b in lo:hi) {
    if (!(b %in% occupied)) {
      if (run == 0L) {
        run_start <- b
      }
      run <- run + 1L
      if (run > best_run) {
        best_run <- run
        center <- (run_start + b) / 2 * bucket + bucket / 2
      }
    } else {
      run <- 0L
    }
  }
  if (best_run < min_gutter_buckets) NA_real_ else center
}

# Render the requested pages of a PDF to PNG images in a child process, run OCR
# on them (in parallel when worthwhile), and return a character vector of text
# (one element per page, in the same order as `pages`).
.run_ocr_pages <- function(input, pages, dpi, psm, preprocess, cores, timeout) {
  # Render page images in a child process (pdf_convert is also Poppler and can
  # crash the same way). The child renders into a temp dir and returns absolute paths.
  render_dir <- tempfile("ocr_pages_")
  dir.create(render_dir, showWarnings = FALSE)
  on.exit(unlink(render_dir, recursive = TRUE), add = TRUE)

  images <- tryCatch(
    callr::r(
      func = function(pdf, pages, dpi, dir) {
        wd <- setwd(dir)
        on.exit(setwd(wd), add = TRUE)
        files <- pdftools::pdf_convert(
          pdf = pdf,
          pages = pages,
          dpi = dpi,
          format = "png",
          verbose = FALSE
        )
        normalizePath(files, winslash = "/", mustWork = FALSE)
      },
      args = list(pdf = input, pages = pages, dpi = dpi, dir = render_dir),
      timeout = timeout
    ),
    error = function(e) NULL
  )
  if (is.null(images)) {
    stop(
      "Could not render PDF pages for OCR (the renderer crashed or timed out on this file): ",
      sQuote(basename(input)),
      call. = FALSE
    )
  }

  # OCR worker: transcribes one page image with a pre-built engine
  ocr_page <- function(img, engine, preprocess) {
    if (preprocess) {
      clean <- magick::image_read(img) |>
        magick::image_convert(colorspace = "gray") |>
        magick::image_deskew() |>
        magick::image_despeckle() |>
        magick::image_normalize()
      img <- magick::image_write(clean, format = "png")
    }
    tesseract::ocr(img, engine = engine)
  }

  # helper to build an engine with the LSTM engine mode
  build_engine <- function(psm) {
    tesseract::tesseract(
      options = list(
        "tessedit_pageseg_mode" = psm,
        "tessedit_ocr_engine_mode" = 1 # LSTM only
      )
    )
  }

  # Parallelize across pages only when it is worth the worker-startup cost.
  use_parallel <- cores > 1L && length(images) >= cores * 2L

  if (use_parallel) {
    cl <- parallel::makeCluster(cores, rscript_args = c("--no-site-file", "--no-init-file"))
    on.exit(parallel::stopCluster(cl), add = TRUE)

    parallel::clusterExport(cl, "build_engine", envir = environment())
    parallel::clusterExport(cl, "psm", envir = environment())
    parallel::clusterEvalQ(cl, {
      .ocr_engine <- build_engine(psm)
      NULL
    })

    # load-balanced: per-page OCR time varies, so dynamic scheduling avoids idle cores
    ocr_text <- parallel::parLapplyLB(
      cl,
      images,
      fun = function(img, preprocess) {
        if (preprocess) {
          clean <- magick::image_read(img) |>
            magick::image_convert(colorspace = "gray") |>
            magick::image_deskew() |>
            magick::image_despeckle() |>
            magick::image_normalize()
          img <- magick::image_write(clean, format = "png")
        }
        tesseract::ocr(img, engine = .ocr_engine)
      },
      preprocess = preprocess
    )
  } else {
    engine <- build_engine(psm)
    ocr_text <- lapply(images, FUN = ocr_page, engine = engine, preprocess = preprocess)
  }

  unlist(ocr_text) |> suppressWarnings()
}

# setup tinytest for checkmate functionality
library("tinytest")
library("checkmate")
using("checkmate")

has_deps_text <- requireNamespace("pdftools", quietly = TRUE) &&
  requireNamespace("tesseract", quietly = TRUE) &&
  requireNamespace("callr", quietly = TRUE)

has_deps_image <- requireNamespace("pdftools", quietly = TRUE) &&
  requireNamespace("magick", quietly = TRUE) &&
  requireNamespace("callr", quietly = TRUE)

pdf <- system.file("extdata", "test.pdf", package = "inlpubs")

# ------ get_pdf_text ------

if (has_deps_text) {
  # returns the output path invisibly
  out <- get_pdf_text(pdf)
  expect_true(is.character(out))

  # output file exists and has a .txt extension
  expect_true(file.exists(out))
  expect_true(grepl("\\.txt$", out, ignore.case = TRUE))

  # file contains non-empty text
  txt <- readLines(out, warn = FALSE)
  expect_true(length(txt) > 0L)
  expect_true(any(nzchar(txt)))

  # text is ASCII only (no non-ASCII characters after normalization)
  expect_true(all(txt == iconv(txt, "UTF-8", "ASCII", sub = "?")))

  # custom output path is respected
  custom <- tempfile(fileext = ".txt")
  get_pdf_text(pdf, output = custom)
  expect_true(file.exists(custom))

  # path without .txt extension gets .txt appended
  no_ext <- tempfile()
  result <- get_pdf_text(pdf, output = no_ext)
  expect_true(grepl("\\.txt$", result, ignore.case = TRUE))

  # invalid input raises an error
  expect_error(get_pdf_text("nonexistent.pdf"))

  # invalid dpi raises an error
  expect_error(get_pdf_text(pdf, dpi = 50L))
  expect_error(get_pdf_text(pdf, dpi = 2000L))

  unlink(c(out, custom, result))
}

# ------ get_pdf_image ------

if (has_deps_image) {
  # returns the output path invisibly
  img <- get_pdf_image(pdf)
  expect_true(is.character(img))

  # output file exists and has a .jpg extension
  expect_true(file.exists(img))
  expect_true(grepl("\\.jpg$", img, ignore.case = TRUE))

  # file is a non-trivially sized JPEG (at least 1 KB)
  expect_true(file.size(img) > 1024L)

  # custom output path is respected
  custom_img <- tempfile(fileext = ".jpg")
  get_pdf_image(pdf, output = custom_img)
  expect_true(file.exists(custom_img))

  # path without .jpg extension gets .jpg appended
  no_ext_img <- tempfile()
  result_img <- get_pdf_image(pdf, output = no_ext_img)
  expect_true(grepl("\\.jpg$", result_img, ignore.case = TRUE))

  # larger width produces a larger file
  img_small <- get_pdf_image(pdf, width = 100L)
  img_large <- get_pdf_image(pdf, width = 500L)
  expect_true(file.size(img_large) > file.size(img_small))

  # invalid input raises an error
  expect_error(get_pdf_image("nonexistent.pdf"))

  # invalid page raises an error
  expect_error(get_pdf_image(pdf, page = 0L))

  # invalid width raises an error
  expect_error(get_pdf_image(pdf, width = 0L))

  unlink(c(img, custom_img, result_img, img_small, img_large))
}

#' Obtain Image from a PDF Document
#'
#' @description Render a single page of a PDF document to a JPEG image. Typically
#'   used to capture a publication's cover (the first page) as a thumbnail.
#'   Requires that the \pkg{pdftools}, \pkg{magick}, and \pkg{callr} packages are
#'   available.
#'
#' @details The selected page is rendered at a resolution computed from the
#'   requested output width (to avoid rasterizing a needlessly large image),
#'   resized with a high-quality Lanczos filter while preserving aspect ratio,
#'   flattened onto a white background, converted to an sRGB JPEG at the given
#'   color depth, and stripped of metadata before being written.
#'
#'   The Poppler-backed reads and the page rasterization are run in a child
#'   process, so a native crash on a problematic PDF (which cannot be caught with
#'   `tryCatch`) surfaces here as a normal R error instead of taking down the
#'   session or aborting a batch run.
#'
#' @param input 'character' string.
#'   File path to the PDF document.
#'
#' @param output 'character' string.
#'   File path for the JPEG image to write. If the path lacks a `.jpg` or `.jpeg`
#'   extension, `.jpg` is appended. Defaults to a temporary file.
#'
#' @param page 'integer' number.
#'   Page number to render. Defaults to 1 (the cover page).
#'
#' @param width 'integer' number.
#'   Output image width in pixels. Height is set automatically to preserve the
#'   page's aspect ratio. Defaults to 300.
#'
#' @param depth 'integer' number.
#'   Image color depth, either 8 or 16. Defaults to 8.
#'
#' @param timeout 'numeric' number.
#'   Maximum time, in seconds, to allow the PDF page to render before the operation is aborted.
#'
#' @param quality 'integer' number between 0 and 100.
#'   JPEG compression quality. Defaults to 70.
#'
#' @return Returns the path to the image file, invisibly.
#'
#' @author J.C. Fisher, U.S. Geological Survey, Idaho Water Science Center
#'
#' @export
#'
#' @seealso [`add_content`] to add cover images to the \pkg{inlpubs} package.
#'
#' @examples
#' input <- system.file("extdata", "test.pdf", package = "inlpubs")
#' path <- get_pdf_image(input)
#'
#' unlink(path)

get_pdf_image <- function(
  input,
  output = tempfile(fileext = ".jpg"),
  page = 1,
  width = 300,
  depth = 8,
  quality = 70,
  timeout = 120
) {
  # check system dependencies
  if (!requireNamespace("pdftools", quietly = TRUE)) {
    stop("Reading a PDF requires the 'pdftools' package", call. = FALSE)
  }
  if (!requireNamespace("magick", quietly = TRUE)) {
    stop("Image processing requires the 'magick' package", call. = FALSE)
  }
  if (!requireNamespace("callr", quietly = TRUE)) {
    stop("Crash-safe PDF reading requires the 'callr' package", call. = FALSE)
  }

  # check arguments
  input <- path.expand(input) |> normalizePath(winslash = "/", mustWork = FALSE)
  checkmate::assert_file_exists(input, access = "r", extension = "pdf")
  output <- path.expand(output) |> normalizePath(winslash = "/", mustWork = FALSE)
  checkmate::assert_path_for_output(output, overwrite = TRUE)
  checkmate::assert_int(page, lower = 1)
  checkmate::assert_int(width, lower = 1)
  checkmate::assert_subset(depth, choices = c(8, 16))
  checkmate::assert_int(quality, lower = 0, upper = 100)
  checkmate::assert_number(timeout, lower = 1)

  # ensure the output has a JPEG file extension
  if (!grepl("\\.(jpg|jpeg)$", tolower(output))) {
    output <- paste0(output, ".jpg")
  }

  # Run all Poppler work in a child process: read the page count and the page's
  # width, then rasterize the requested page to a PNG file. A magick image object
  # cannot cross the process boundary (it holds an external pointer), so the child
  # writes a PNG and returns its path; the parent does the safe post-processing.
  # A native crash on a problematic PDF is caught here rather than killing R.
  render_dir <- tempfile("pdf_image_")
  dir.create(render_dir, showWarnings = FALSE)
  on.exit(unlink(render_dir, recursive = TRUE), add = TRUE)

  rendered <- tryCatch(
    callr::r(
      func = function(pdf, page, width, dir) {
        # validate the page exists within the child (uses Poppler metadata)
        n_pages <- pdftools::pdf_info(pdf)$pages
        if (page < 1 || page > n_pages) {
          stop(sprintf("page %d out of range (document has %d)", page, n_pages))
        }

        # compute a render resolution sized to the target width rather than
        # rasterizing the page at full resolution and discarding most of it.
        # Page width is in points; inches = points / 72; dpi = target px / inches.
        # A 1.5x headroom keeps the later resize a mild downscale. Falls back to
        # 150 dpi if the page width is missing or invalid, and is clamped.
        page_width_pt <- pdftools::pdf_pagesize(pdf)$width[page]
        if (is.na(page_width_pt) || page_width_pt <= 0) {
          render_dpi <- 150L
        } else {
          page_width_in <- page_width_pt / 72
          render_dpi <- ceiling((width * 1.5) / page_width_in)
          render_dpi <- as.integer(max(72L, min(render_dpi, 300L)))
        }

        # rasterize just the requested page to a PNG (Poppler via magick)
        png_path <- file.path(dir, "page.png")
        magick::image_read_pdf(pdf, pages = page, density = render_dpi) |>
          magick::image_write(path = png_path, format = "png")

        list(png = normalizePath(png_path, winslash = "/", mustWork = FALSE))
      },
      args = list(pdf = input, page = page, width = width, dir = render_dir),
      timeout = timeout
    ),
    error = function(e) NULL
  )
  if (is.null(rendered)) {
    stop(
      "Could not render PDF page (the renderer crashed or timed out on this file): ",
      sQuote(basename(input)),
      call. = FALSE
    )
  }

  # post-process the rendered PNG in the parent (no Poppler involved here):
  #   resize to width with a high-quality filter (aspect ratio preserved),
  #   flatten onto white to avoid transparency rendering as black in JPEG,
  #   convert to sRGB at the requested depth, then strip metadata
  image <- magick::image_read(rendered$png) |>
    magick::image_resize(
      geometry = magick::geometry_size_pixels(width = width),
      filter = "Lanczos"
    ) |>
    magick::image_background(color = "white") |>
    magick::image_flatten() |>
    magick::image_convert(
      colorspace = "sRGB",
      depth = depth
    ) |>
    magick::image_strip()

  # write image as JPEG at the requested quality
  magick::image_write(image, path = output, format = "jpeg", quality = quality)

  # confirm the file was written
  checkmate::assert_file_exists(output, access = "r")

  invisible(output)
}

#' Term Frequency from INLPO Publications
#'
#' @description A term-frequency table giving the number of times each term appears in
#'   each publication, for publications by the U.S. Geological Survey (USGS), Idaho
#'   Water Science Center, Idaho National Laboratory Project Office (INLPO). The table
#'   has one row per term-publication pairing.
#'
#' @format An object of class 'term' that inherits from the 'data.frame' class and
#'   includes the following columns:
#'   \describe{
#'     \item{`term`}{Term (a single word, or a multi-word n-gram), stored as a factor.
#'       Terms are lowercase ASCII.}
#'     \item{`pub_id`}{Identifier of the publication in which the term appears, stored
#'       as a factor; a foreign key referencing the `pub_id` primary key of the
#'       [`pubs`] table.}
#'     \item{`freq`}{Number of times the term occurs in that publication.}
#'   }
#'
#' @source The publication text was sourced from the original PDF documents using the
#'   [`get_pdf_text`] function, and term frequencies were extracted from the text
#'   using the [`mine_text`] function.
#'
#' @seealso [`mine_text`] to build a term-frequency table, [`search_terms`] to search
#'   within it, and [`make_wordcloud`] to visualize it.
#'
#' @keywords datasets
#'
#' @examples
#' str(terms, max.level = 3, width = 75, strict.width = "cut")
#'
#' # Most frequent terms across all publications:
#' freq_by_term <- tapply(terms$freq, terms$term, FUN = sum)
#' sort(freq_by_term, decreasing = TRUE) |> head(n = 10)

"terms"

# setup tinytest for checkmate functionality
library("tinytest")
library("checkmate")
using("checkmate")

# ------ pubs ------

# class
expect_true(inherits(pubs, "pub"))
expect_true(inherits(pubs, "data.frame"))

# dimensions and columns
expect_true(nrow(pubs) > 0L)
expect_equal(
  colnames(pubs),
  c(
    "pub_id",
    "institution",
    "type",
    "text_ref",
    "year",
    "author_id",
    "title",
    "bibentry",
    "abstract",
    "annotation",
    "annotation_src",
    "files"
  )
)

# row names are the primary key and are unique
expect_true(length(unique(row.names(pubs))) == nrow(pubs))

# pub_id column matches row names
expect_equal(pubs[["pub_id"]], row.names(pubs))

# year is a positive integer within a sensible range
expect_true(is.integer(pubs[["year"]]))
expect_true(all(pubs[["year"]] >= 1949L))

# list column names are aligned with row names
expect_equal(names(pubs[["author_id"]]), row.names(pubs))
expect_equal(names(pubs[["bibentry"]]), row.names(pubs))
expect_equal(names(pubs[["files"]]), row.names(pubs))

# every publication has at least one author
expect_true(all(lengths(pubs[["author_id"]]) >= 1L))

# annotation_src values, when present, reference a valid pub_id
expect_true(
  all(is.na(pubs[["annotation_src"]]) | pubs[["annotation_src"]] %in% row.names(pubs))
)

# ------ authors ------

# class
expect_true(inherits(authors, "author"))
expect_true(inherits(authors, "data.frame"))

# dimensions and columns
expect_true(nrow(authors) > 0L)
expect_equal(
  colnames(authors),
  c(
    "author_id",
    "name",
    "person",
    "pub_id",
    "total_pub",
    "single_authored",
    "multi_authored",
    "first_authored",
    "first_year",
    "last_year"
  )
)

# row names are the primary key and are unique
expect_true(length(unique(row.names(authors))) == nrow(authors))

# list column names are aligned with row names
expect_equal(names(authors[["person"]]), row.names(authors))
expect_equal(names(authors[["pub_id"]]), row.names(authors))

# every author has at least one publication
expect_true(all(authors[["total_pub"]] >= 1L))

# publication count breakdown is consistent
expect_equal(
  authors[["single_authored"]] + authors[["multi_authored"]],
  authors[["total_pub"]]
)

# year range is sensible
expect_true(all(authors[["first_year"]] <= authors[["last_year"]]))

# ------ terms ------

# class
expect_true(inherits(terms, "term"))
expect_true(inherits(terms, "data.frame"))

# columns
expect_equal(colnames(terms), c("term", "pub_id", "freq"))

# term and pub_id are factors, freq is integer
expect_true(is.factor(terms[["term"]]))
expect_true(is.factor(terms[["pub_id"]]))
expect_true(is.integer(terms[["freq"]]))

# no zero- or negative-frequency rows
expect_true(all(terms[["freq"]] > 0L))

# terms are lowercase
expect_equal(
  as.character(terms[["term"]]),
  tolower(as.character(terms[["term"]]))
)

# ------ cross-dataset referential integrity ------

# every author_id in pubs references a valid row in authors
expect_true(all(unlist(pubs[["author_id"]]) %in% row.names(authors)))

# every pub_id in authors references a valid row in pubs
expect_true(all(unlist(authors[["pub_id"]]) %in% row.names(pubs)))

# every pub_id level in terms references a valid row in pubs
expect_true(all(levels(terms[["pub_id"]]) %in% row.names(pubs)))

# ------ existing test (annotation sources) ------

# every annotated publication has a non-missing annotation_src
is <- pubs[["annotation_src"]][!is.na(pubs[["annotation"]])] |> anyNA()
expect_false(is)

```{js, echo=FALSE}

// Custom range filter with value label
// Code copied from R-package 'reactable' vignette at
// https://github.com/glin/reactable/blob/HEAD/vignettes/custom-filtering.Rmd
// Accessed on 2024-02-12
// License: MIT
// YEAR: 2019
// COPYRIGHT HOLDER: Greg Lin, Tanner Linsley

function rangeFilter(column, state) {
  // Get min and max values from raw table data
  let min = Infinity;
  let max = 0;
  state.data.forEach(function (row) {
    const value = row[column.id];
    if (value < min) {
      min = Math.floor(value);
    } else if (value > max) {
      max = Math.ceil(value);
    }
  });

  const filterValue = column.filterValue || min;
  const input = React.createElement("input", {
    type: "range",
    value: filterValue,
    min: min,
    max: max,
    onChange: function (event) {
      // Set to undefined to clear the filter
      column.setFilter(event.target.value || undefined);
    },
    style: { width: "100%", marginRight: "8px" },
    "aria-label": "Filter " + column.name,
  });

  return React.createElement(
    "div",
    { style: { display: "flex", alignItems: "center", height: "100%" } },
    [input, filterValue]
  );
}

// Filter method that filters numeric columns by minimum value
function filterMinValue(rows, columnId, filterValue) {
  return rows.filter(function (row) {
    return row.values[columnId] >= filterValue;
  });
}

// Filter method that filters numeric columns by maximum value
function filterMaxValue(rows, columnId, filterValue) {
  return rows.filter(function (row) {
    return row.values[columnId] <= filterValue;
  });
}

```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_person.R
\name{get_person}
\alias{get_person}
\title{Select Persons by Identifier}
\usage{
get_person(x, persons)
}
\arguments{
\item{x}{'character' vector.
Identifier(s) of the person(s) to select. Must be non-missing and unique, and
each must match a name in \code{persons}.}

\item{persons}{'person' named list.
Information about an arbitrary number of persons, as produced by
\code{\link[utils:person]{utils::person()}}. Each element is named with the identifier that uniquely
identifies that person.}
}
\value{
A 'person' list containing the requested subset of \code{persons}, in the
order given by \code{x}.
}
\description{
Select one or more persons from a named list by their unique
identifiers, preserving the requested order. All identifiers must be present;
any that are missing raise an error.
}
\examples{
get_person("jfisher", persons = inlpubs::authors$person)
}
\author{
J.C. Fisher, U.S. Geological Survey, Idaho Water Science Center
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_content.R
\name{add_content}
\alias{add_content}
\title{Add Content from PDF Documents}
\usage{
add_content(
  pub_id,
  year,
  type = c("text", "image"),
  ...,
  srcdir = "archive",
  destdir = tempdir(),
  ignore = NULL,
  pubs = inlpubs::pubs,
  overwrite = FALSE
)
}
\arguments{
\item{pub_id}{'character' vector.
Unique identifier(s) of the publication(s) to process.
Publications may instead, or additionally, be selected with \code{year}.}

\item{year}{'integer' vector.
Year(s) of publication. All publications from the given year(s) are selected,
in addition to any specified by \code{pub_id}.}

\item{type}{'character' string.
Type of content to extract from each PDF, either \code{"text"} (the default) or
\code{"image"} (a cover image).}

\item{...}{Additional arguments passed to the extraction function:
\code{\link{get_pdf_text}} when \code{type = "text"}, or \code{\link{get_pdf_image}} when \code{type = "image"}.}

\item{srcdir}{'character' string.
Source directory containing the publication PDFs, organized into
year subdirectories (see Details). Defaults to \code{"archive"} in the working directory.}

\item{destdir}{'character' string.
Destination directory for the extracted output files. Created if it does not
already exist. Defaults to the session's temporary directory.}

\item{ignore}{'character' vector.
Publication identifier(s) to skip. Useful for documents that are known to fail
automatic extraction.}

\item{pubs}{'pub' table.
Publications of the INLPO; see the \code{\link{pubs}} dataset for the expected data format.}

\item{overwrite}{'logical' flag.
Whether to overwrite an existing output file. Defaults to \code{FALSE}, in which case
publications with an existing output are skipped.}
}
\value{
Invisibly returns a 'character' vector of paths to the saved output files,
carrying a \code{"failed"} attribute with the identifiers of any publications whose
extraction failed. Returns \code{NULL} (invisibly) when no publications remain to
process.
}
\description{
Extract the text or cover image from publication PDF documents and
save the result to disk, for incorporation into the \pkg{inlpubs} package corpus.
Publications are selected by identifier or by year, and each must already be
present in the \code{pubs} dataset before this function is run.
}
\details{
For each selected publication, the function locates its primary PDF (the
first file listed in the publication's \code{files} entry) within the source directory,
then delegates extraction to \code{\link{get_pdf_text}} (for \code{type = "text"}) or
\code{\link{get_pdf_image}} (for \code{type = "image"}). Extra arguments passed through \code{...}
are forwarded to whichever of those functions is used.

Source documents are expected to follow the archive layout
\verb{srcdir/<year>/<pub_id>/<file>}, for example
\code{archive/2005/KnobelOthers2005/ofr20051223.pdf}. Output files are written to
\code{destdir} and named \verb{pub-<pub_id>}, with the appropriate extension (\code{.txt} or
\code{.jpg}) supplied by the extraction function.

Publications whose \code{files} entry is empty, and those listed in \code{ignore}, are
skipped with a message. By default, existing output files are left in place;
set \code{overwrite = TRUE} to regenerate them.

Extraction failures (including renderer crashes or timeouts on problematic
PDFs) are caught per publication: the offending document is reported and
skipped, and processing continues with the next one. The identifiers of any
failed documents are returned in the \code{"failed"} attribute of the result.

Processing the full archive can take a considerable amount of time. For routine
use, run this function interactively and one publication year at a time so that
progress can be monitored.
}
\examples{
\dontrun{
  # Extract text for a single publication into the corpus folder
  add_content("KnobelOthers2005", destdir = "data-raw/corpus")

  # Extract text for every publication from a given year
  add_content(year = 2005, destdir = "data-raw/corpus")

  # Extract cover images into the vignettes folder
  add_content(year = 2005, type = "image", destdir = "vignettes")

  # Process several years, skipping a document that fails extraction
  add_content(year = 2003:2005, ignore = "Mazurek2004", destdir = "data-raw/corpus")
}
}
\seealso{
\code{\link{get_pdf_text}} and \code{\link{get_pdf_image}} for the underlying extraction
functions, and the \code{\link{pubs}} dataset for publication metadata.
}
\author{
J.C. Fisher, U.S. Geological Survey, Idaho Water Science Center
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_pdf_image.R
\name{get_pdf_image}
\alias{get_pdf_image}
\title{Obtain Image from a PDF Document}
\usage{
get_pdf_image(
  input,
  output = tempfile(fileext = ".jpg"),
  page = 1,
  width = 300,
  depth = 8,
  quality = 70,
  timeout = 120
)
}
\arguments{
\item{input}{'character' string.
File path to the PDF document.}

\item{output}{'character' string.
File path for the JPEG image to write. If the path lacks a \code{.jpg} or \code{.jpeg}
extension, \code{.jpg} is appended. Defaults to a temporary file.}

\item{page}{'integer' number.
Page number to render. Defaults to 1 (the cover page).}

\item{width}{'integer' number.
Output image width in pixels. Height is set automatically to preserve the
page's aspect ratio. Defaults to 300.}

\item{depth}{'integer' number.
Image color depth, either 8 or 16. Defaults to 8.}

\item{quality}{'integer' number between 0 and 100.
JPEG compression quality. Defaults to 70.}

\item{timeout}{'numeric' number.
Maximum time, in seconds, to allow the PDF page to render before the operation is aborted.}
}
\value{
Returns the path to the image file, invisibly.
}
\description{
Render a single page of a PDF document to a JPEG image. Typically
used to capture a publication's cover (the first page) as a thumbnail.
Requires that the \pkg{pdftools}, \pkg{magick}, and \pkg{callr} packages are
available.
}
\details{
The selected page is rendered at a resolution computed from the
requested output width (to avoid rasterizing a needlessly large image),
resized with a high-quality Lanczos filter while preserving aspect ratio,
flattened onto a white background, converted to an sRGB JPEG at the given
color depth, and stripped of metadata before being written.

The Poppler-backed reads and the page rasterization are run in a child
process, so a native crash on a problematic PDF (which cannot be caught with
\code{tryCatch}) surfaces here as a normal R error instead of taking down the
session or aborting a batch run.
}
\examples{
input <- system.file("extdata", "test.pdf", package = "inlpubs")
path <- get_pdf_image(input)

unlink(path)
}
\seealso{
\code{\link{add_content}} to add cover images to the \pkg{inlpubs} package.
}
\author{
J.C. Fisher, U.S. Geological Survey, Idaho Water Science Center
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/search_terms.R
\name{search_terms}
\alias{search_terms}
\title{Search a Term-Frequency Table}
\usage{
search_terms(
  x,
  data = inlpubs::terms,
  ignore_case = TRUE,
  ...,
  low_freq = 1,
  high_freq = Inf,
  simplify = TRUE
)
}
\arguments{
\item{x}{'character' string.
Regular-expression pattern to search for among the table's terms.}

\item{data}{'term' and 'data.frame' class.
Term-frequency table to search, with columns \code{term}, \code{pub_id}, and \code{freq}.
Defaults to the term frequencies from the INLPO publications; see the \code{\link{terms}}
dataset for details.}

\item{ignore_case}{'logical' flag.
Whether to ignore character case when matching. Defaults to \code{TRUE}.}

\item{...}{Additional arguments passed to \code{\link[=grep]{grep()}} (for example, \code{fixed} or \code{perl}).}

\item{low_freq}{'numeric' number.
Lower frequency bound; terms occurring fewer than this many times are excluded.
Defaults to 1.}

\item{high_freq}{'numeric' number.
Upper frequency bound; terms occurring more than this many times are excluded.
Defaults to \code{Inf} (no upper limit).}

\item{simplify}{'logical' flag.
Whether to return only the unique publication identifiers (\code{TRUE}, the default)
rather than the full table of matching rows.}
}
\value{
If \code{simplify = TRUE}, a 'character' vector of unique publication
identifiers, ordered by the most frequent matching term first. If
\code{simplify = FALSE}, a 'data.frame' of matching rows sorted by decreasing
frequency.
}
\description{
Search a term-frequency table by pattern or frequency.
}
\examples{
search_terms("mlms")

out <- search_terms("mlms", simplify = FALSE)
head(out)
}
\seealso{
\code{\link{mine_text}} to build a term-frequency table.
}
\author{
J.C. Fisher, U.S. Geological Survey, Idaho Water Science Center
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pubs.R
\docType{data}
\name{pubs}
\alias{pubs}
\title{Bibliographic Records for INLPO Publications}
\format{
An object of class 'pub' that inherits from the 'data.frame' class and
includes the following columns:
\describe{
\item{\code{pub_id}}{Unique identifier for the publication (the table's primary
key), in upper camel case (for example, \code{"FisherOthers2012"}).}
\item{\code{institution}}{Name of the institution that published and/or sponsored
the report.}
\item{\code{type}}{Type of publication, such as \code{"Open-File Report"},
\code{"Data Series"}, or \code{"Manual"}.}
\item{\code{text_ref}}{Text reference (the in-text citation) excluding the year of
publication, such as \code{"Dietz"} or \code{"Treinen and others"}.}
\item{\code{year}}{Year of publication.}
\item{\code{author_id}}{A named list of character vectors giving the identifier(s)
of the contributing author(s), in author order; each value is a foreign key
referencing the \code{author_id} primary key of the \code{\link{authors}} table. List names
match \code{pub_id}.}
\item{\code{title}}{Title of publication.}
\item{\code{bibentry}}{A named list of \link[utils:bibentry]{bibentry} objects holding
the full bibliographic record, from which formatted citations and BibTeX
entries are derived. List names match \code{pub_id}.}
\item{\code{abstract}}{Abstract of publication, or \code{NA} when unavailable.}
\item{\code{annotation}}{Annotation summarizing the publication, or \code{NA} when none
is provided.}
\item{\code{annotation_src}}{Identifier (\code{pub_id}) of the publication the annotation
was taken from, or \code{NA} when not applicable
(Knobel and others, 2005; Bartholomay, 2022).}
\item{\code{files}}{A named list of character vectors naming the file(s) associated
with the publication, primary document first, or \code{NULL} when no files are
recorded. List names match \code{pub_id}.}
}
}
\source{
Many of these publications are available through the
\href{https://pubs.usgs.gov/}{USGS Publications Warehouse}.
}
\usage{
pubs
}
\description{
Bibliographic information for reports, articles, maps, and theses
related to scientific monitoring and research conducted by the U.S. Geological
Survey (USGS), Idaho Water Science Center, Idaho National Laboratory Project
Office (INLPO).
}
\examples{
# Subset Fisher and others (2012) and display its structure:
id <- "FisherOthers2012"
pub <- pubs[id, ]
str(pub, max.level = 3, width = 75, strict.width = "cut")

# Print the suggested citation:
attr(unclass(pub$bibentry[[1]])[[1]], which = "textVersion")

# Print the authors' full names:
format(pub$bibentry[[1]]$author, include = c("given", "family"))

# Print the abstract:
pub$abstract
}
\references{
Bartholomay, R.C., 2022, Historical development of the U.S. Geological Survey
hydrological monitoring and investigative programs at the Idaho National Laboratory, Idaho, 2002-2020:
U.S. Geological Survey Open-File Report 2022-1027 (DOE/ID-22256), 54 p., \doi{10.3133/ofr20221027}.

Knobel, L.L., Bartholomay, R.C., and Rousseau, J.P., 2005,
Historical development of the U.S. Geological Survey hydrologic monitoring and investigative programs
at the Idaho National Engineering and Environmental Laboratory, Idaho, 1949 to 2001:
U.S. Geological Survey Open-File Report 2005--1223 (DOE/ID--22195), 93 p.,
\doi{10.3133/ofr20051223}.
}
\seealso{
\code{\link{authors}} for the contributing authors referenced by \code{author_id},
\code{\link{terms}} for term frequencies derived from publication text, and
\code{\link{search_terms}} to query publications by term.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_pdf_text.R
\name{get_pdf_text}
\alias{get_pdf_text}
\title{Obtain Text from a PDF Document}
\usage{
get_pdf_text(
  input,
  output = tempfile(fileext = ".txt"),
  dpi = 300,
  psm = 1,
  force_ocr = FALSE,
  preprocess = FALSE,
  min_chars = 20,
  cores = NULL,
  timeout = 120
)
}
\arguments{
\item{input}{'character' string.
File path to the PDF document.}

\item{output}{'character' string.
File path for the text file to write. If the path lacks a \code{.txt} or \code{.text}
extension, \code{.txt} is appended. Defaults to a temporary file.}

\item{dpi}{'integer' number between 100 and 1200.
Resolution, in dots per inch, at which pages are rendered before OCR.
Higher values improve OCR accuracy at the cost of speed; 300 (the default)
suits most documents, while 400 to 600 may help with small fonts or poor
scans. Ignored for pages read directly from the text layer.}

\item{psm}{'integer' number between 0 and 13.
Tesseract Page Segmentation Mode, describing the expected page layout.
The default of 1 segments the page automatically and detects orientation and
script, which suits multi-column documents.}

\item{force_ocr}{'logical' flag.
Whether to OCR every page regardless of its text layer. Use this when the
embedded text is known to be unreliable (e.g. fonts that extract as gibberish,
which \code{min_chars} cannot catch because the gibberish is long enough to pass).
Defaults to \code{FALSE}.}

\item{preprocess}{'logical' flag.
Whether to grayscale, deskew, despeckle, and normalize each page image before
OCR, which improves accuracy on scanned documents at a modest per-page cost.
Requires the \pkg{magick} package. Applies only to pages that are OCR'd.
Defaults to \code{FALSE}.}

\item{min_chars}{'integer' number.
Minimum number of non-whitespace characters a page's extracted text must
have to be accepted without OCR. Pages below this threshold are sent to OCR.
Defaults to 20.}

\item{cores}{'integer' number.
Number of parallel workers to use for OCR. When \code{NULL} (the default), the
count is taken from the \code{OCR_CORES} environment variable if set (useful for
honoring a container CPU quota), otherwise from one fewer than the number of
detected cores.}

\item{timeout}{'numeric' number.
Maximum time, in seconds, to allow each child-process call to run before it
is aborted. Applied independently to the text-extraction call and the
page-rendering call. Defaults to 120.}
}
\value{
Returns the path to the text file, invisibly.
Each page of the PDF is written as a separate line.
}
\description{
Extract text from a PDF document and write it to a plain-text
file, one page per line. Pages with an extractable text layer are read
directly; pages without one (e.g. scanned images) fall back to optical
character recognition (OCR). The text is normalized to ASCII before writing.

Requires the \pkg{pdftools}, \pkg{tesseract}, and \pkg{callr} packages;
image preprocessing additionally requires \pkg{magick}.
}
\details{
Direct extraction is attempted first, as it is faster and more accurate than
OCR for born-digital PDFs; a page is OCR'd only when its text layer yields
fewer than \code{min_chars} non-whitespace characters, or when \code{force_ocr} is
\code{TRUE}. Poppler reads and page rendering run in isolated child processes, so
a renderer crash surfaces as a normal R error rather than aborting the
session.

Directly extracted text is reordered into human reading order from word
bounding boxes: full-width headers and captions are set aside, then the body
is split into columns only on strong evidence of a multi-column layout. OCR
uses Tesseract's LSTM engine, run in parallel across \code{cores} workers when
several pages need it. The result is normalized to strictly ASCII: selected
scientific symbols become ASCII tokens, accents are transliterated, line-break
hyphenation is rejoined, and quotes, control characters, and repeated
punctuation and whitespace are cleaned up.
}
\examples{
\dontrun{
  input <- system.file("extdata", "test.pdf", package = "inlpubs")
  path <- get_pdf_text(input)

  unlink(path)
}
}
\seealso{
\code{\link{add_content}} to add extracted text to the \pkg{inlpubs}-package corpus.
}
\author{
J.C. Fisher, U.S. Geological Survey, Idaho Water Science Center
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_pubs.R
\name{plot_pubs}
\alias{plot_pubs}
\title{Plot Annual Publication Counts}
\usage{
plot_pubs(
  pubs = inlpubs::pubs,
  author_id = NULL,
  full_range = TRUE,
  legend_loc = "topleft",
  colors = c(USGS = "#4682B4", `Non-USGS` = "#B47846"),
  ...
)
}
\arguments{
\item{pubs}{'pub' data frame.
Publications of the INLPO; see the \code{\link{pubs}} dataset for the expected data
format. Defaults to \code{inlpubs::pubs}.}

\item{author_id}{'character' vector, or \code{NULL}.
One or more author identifiers used to subset \code{pubs} to publications that
any of those authors appear on before plotting. The default (\code{NULL})
includes all publications (no author filtering).}

\item{full_range}{'logical' flag.
Whether the x-axis spans the year range of all publications in \code{pubs}
(before any author filtering). If \code{TRUE} (the default), the timeline is
fixed to the full dataset, making per-author charts directly comparable.
If \code{FALSE}, the timeline is limited to the year range of the filtered
publications. Has no effect when \code{author_id} is \code{NULL}.}

\item{legend_loc}{'character' string.
Position of the legend, passed as the keyword location to
\code{\link[graphics:legend]{graphics::legend()}}. One of \code{"topleft"} (the default), \code{"topright"},
\code{"bottomleft"}, \code{"bottomright"}, \code{"top"}, \code{"bottom"}, \code{"left"},
\code{"right"}, or \code{"center"}.}

\item{colors}{'character' vector of length 2.
Fill colors for the two series, in the order \code{c("USGS", "Non-USGS")}.
Defaults to steel blue and burnt orange.}

\item{...}{Additional arguments passed to \code{\link[graphics:barplot]{graphics::barplot()}}.}
}
\value{
Invisibly returns a numeric vector giving the x-axis coordinates of
the bar midpoints, as returned by \code{\link[graphics:barplot]{graphics::barplot()}}.
}
\description{
Draw a stacked bar chart of INLPO publication counts by year,
with USGS and non-USGS publications distinguished by color.
}
\examples{
plot_pubs()
}
\seealso{
\code{\link{pubs}} for the publications dataset.
}
\author{
J.C. Fisher, U.S. Geological Survey, Idaho Water Science Center
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filter_data_list.R
\name{filter_data_list}
\alias{filter_data_list}
\title{Filter Data List Column}
\usage{
filter_data_list(table_id, style = "width: 100\%; height: 28px;")
}
\arguments{
\item{table_id}{'character' string.
Unique identifier of the table the filter belongs to. Must match the \code{elementId}
given to the table so that \code{Reactable.setFilter()} can target it.}

\item{style}{'character' string.
CSS applied to the text input element. Defaults to a full-width, 28px-tall box.}
}
\value{
Returns a function of two arguments, \code{values} (the column's values) and
\code{name} (the column name), which returns an \pkg{htmltools} tag list for the
filter input.
}
\description{
Create a column filter for a \pkg{reactable} table that pairs a text
input with an HTML \code{datalist}, giving users autocomplete suggestions drawn from
the column's unique values.
Requires that the \pkg{htmltools} package is available.
}
\details{
The returned function is intended for use as a column's \code{filterInput} in
\code{\link[reactable:colDef]{reactable::colDef()}}. When rendered, it produces a text box backed by a
\code{datalist} of the sorted unique values in that column; typing filters the table
via \code{Reactable.setFilter()}, while the suggestions assist exact matches.
}
\examples{
f <- filter_data_list("table-id")
}
\references{
Adapted from the \pkg{reactable} "Custom Filtering" vignette
(\url{https://github.com/glin/reactable/blob/HEAD/vignettes/custom-filtering.Rmd}),
by Greg Lin and Tanner Linsley, 2019 (MIT License). Accessed 2024-05-30.
}
\author{
J.C. Fisher, U.S. Geological Survey, Idaho Water Science Center
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mine_text.R
\name{mine_text}
\alias{mine_text}
\title{Build a Term-Frequency Table}
\usage{
mine_text(docs, ngmin = 1, ngmax = ngmin, sparse = NULL)
}
\arguments{
\item{docs}{'list' or 'character' vector.
Document text to analyze. Each element holds the extracted text of a single
document and may itself be a character vector (its pieces are concatenated).
\code{NA} values are dropped. If unnamed, elements are numbered sequentially.}

\item{ngmin, ngmax}{'integer' numbers.
Minimum and maximum number of words per term (n-gram). The default
(\code{ngmin = ngmax = 1}) counts single words. Values greater than one require the
\pkg{RWeka} package and that the \code{JAVA_HOME} environment variable point to a
Java installation. Recommended for a single document only.}

\item{sparse}{'numeric' number greater than 0 and less than 1, or \code{NULL}.
Maximum allowed sparsity: terms appearing in a smaller proportion of documents
are removed. For example, \code{sparse = 0.99} removes terms more sparse than 0.99,
while \code{sparse = 0.01} retains only terms appearing in nearly every document.
The default (\code{NULL}) disables sparse-term removal.}
}
\value{
A 'data.frame' term-frequency table with columns \code{term}, \code{pub_id}, and
\code{freq}, giving the number of times each term occurs in each document. Rows with
a zero count are omitted, and the table is sorted by term (ascending) then
frequency (descending).
}
\description{
Build a term-frequency table from document text.
A term is a single word or, when n-gram tokenization is enabled, an ordered
sequence of words.
}
\details{
Documents are cleaned before counting: URLs, escaped unicode sequences,
HTML tags, and selected punctuation are removed; text is lowercased; numbers,
English and SMART stopwords, and remaining punctuation are stripped; and
whitespace is collapsed. HTML entities are decoded when the \pkg{textutils}
package is available. When n-grams are requested, terms containing a repeated
word are discarded.
}
\examples{
d <- c(
  "The quick brown fox jumps over the lazy lazy dog.",
  "Pack my brown box.",
  "Jazz fly brown dog."
) |>
  mine_text()

d <- list(
  "A" = "The quick brown fox jumps over the lazy lazy dog.",
  "B" = c("Pack my brown box.", NA, "Jazz fly brown dog."),
  "C" = NA_character_
) |>
  mine_text()
}
\seealso{
\code{\link{search_terms}} to search within the resulting table, and
\code{\link{make_wordcloud}} to create a word cloud from it.
}
\author{
J.C. Fisher, U.S. Geological Survey, Idaho Water Science Center
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/make_wordcloud.R
\name{make_wordcloud}
\alias{make_wordcloud}
\title{Create a Word Cloud from Term Frequencies}
\usage{
make_wordcloud(
  x,
  max_terms = 200,
  size = 1,
  shape = "circle",
  ellipticity = 0.65,
  ...,
  width = 910,
  output = NULL,
  display = FALSE
)
}
\arguments{
\item{x}{'data.frame'.
Term-frequency table containing at least the columns \code{term} and \code{freq}.}

\item{max_terms}{'integer' number.
Maximum number of terms to include, taken in order of decreasing frequency.
Defaults to 200.}

\item{size}{'numeric' number.
Font-size scaling factor. Defaults to 1.}

\item{shape}{'character' string.
Shape of the cloud. One of \code{"circle"} (the default), \code{"cardioid"}, \code{"diamond"},
\code{"triangle-forward"}, \code{"triangle"}, \code{"pentagon"}, or \code{"star"}.}

\item{ellipticity}{'numeric' number between 0 and 1.
Degree of flattening applied to the shape. Defaults to 0.65.}

\item{...}{Additional arguments passed to \code{\link[wordcloud2:wordcloud2]{wordcloud2::wordcloud2()}}.}

\item{width}{'integer' number.
Image width in pixels. Defaults to 910. Image height is derived from \code{width}
and \code{ellipticity}.}

\item{output}{'character' string.
Path to the output PNG file. Defaults to a temporary file.}

\item{display}{'logical' flag.
Whether to display the saved PNG in a graphics window. Requires the \pkg{magick}
package. Defaults to \code{FALSE}.}
}
\value{
Invisibly returns the file path to the word-cloud PNG.
}
\description{
Create a word cloud from a term-frequency table and save it as a PNG
file. The most frequent terms are rendered, laid out within the chosen shape.
Requires the \pkg{htmltools}, \pkg{htmlwidgets}, \pkg{magick}, \pkg{webshot2},
and \pkg{wordcloud2} packages. System dependencies are
\href{https://imagemagick.org/}{ImageMagick} (for displaying the PNG),
\href{https://optipng.sourceforge.net/}{OptiPNG} (for PNG compression), and a
\href{https://www.google.com/chrome/}{Chrome} or Chromium-based browser with
support for the Chrome DevTools protocol. Use
\code{\link[chromote:find_chrome]{chromote::find_chrome()}} to locate the browser.
}
\examples{
\dontrun{
  d <- wordcloud2::demoFreq |> head(n = 10)
  colnames(d) <- c("term", "freq")
  file <- make_wordcloud(d, display = interactive())

  unlink(file)
}
}
\seealso{
\code{\link{mine_text}} to build a term-frequency table.
}
\author{
J.C. Fisher, U.S. Geological Survey, Idaho Water Science Center
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/make_webpages.R
\name{make_webpages}
\alias{make_webpages}
\title{Create Author and Publication Webpages}
\usage{
make_webpages(
  authors = NULL,
  pubs = NULL,
  destdir = tempdir(),
  coverdir = "vignettes",
  quiet = FALSE
)
}
\arguments{
\item{authors}{'author' data frame.
Contributing authors to the INLPO publications; see the \code{\link{authors}} dataset for
the expected data format. When \code{NULL}, author webpages are not written.}

\item{pubs}{'pub' data frame.
Publications of the INLPO; see the \code{\link{pubs}} dataset for the expected data format.
When \code{NULL}, publication webpages are not written.}

\item{destdir}{'character' string.
Destination directory for the written files, with tilde-expansion performed.
Created if it does not exist. Defaults to a temporary directory.}

\item{coverdir}{'character' string.
Directory searched for publication cover images named \verb{pub-<id>.jpg}.
Defaults to \code{"vignettes"}.}

\item{quiet}{'logical' flag.
Whether to suppress progress messages. Defaults to \code{FALSE}.}
}
\value{
Invisibly returns \code{NULL}. Called for its side effect of writing files.
}
\description{
Create a webpage for each author and each publication, listing the
author's publications or the publication's details. Each webpage is written as
an \href{https://rmarkdown.rstudio.com/}{R Markdown} (\code{.Rmd}) file.
}
\examples{
destdir <- tempfile("")
make_webpages(
  authors = inlpubs::authors,
  pubs = inlpubs::pubs,
  destdir = destdir,
  quiet = TRUE
)

unlink(destdir, recursive = TRUE)
}
\author{
J.C. Fisher, U.S. Geological Survey, Idaho Water Science Center
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/authors.R
\docType{data}
\name{authors}
\alias{authors}
\title{Contributing Authors to INLPO Publications}
\format{
An object of class 'author' that inherits from the 'data.frame' class and
contains 267 rows and the following 10 columns:
\describe{
\item{\code{author_id}}{Unique identifier for the author, composed of the author's
first initial and last name in lowercase (for example, \code{"jfisher"}).}
\item{\code{name}}{Author name, surname first, followed by initials or given name.}
\item{\code{person}}{A named list of \link[utils:person]{person} objects holding
contact details such as email address and
\href{https://orcid.org/}{ORCID} identifier. Names match \code{author_id}.}
\item{\code{pub_id}}{A named list of character vectors giving the identifier(s) of
the publication(s) the author contributed to; each value is a foreign key
referencing the \code{pub_id} primary key of the \code{\link{pubs}} table.}
\item{\code{total_pub}}{Total number of publications.}
\item{\code{single_authored}}{Number of single-authored publications.}
\item{\code{multi_authored}}{Number of multi-authored publications.}
\item{\code{first_authored}}{Number of multi-authored publications in which the
author appears as the first author.}
\item{\code{first_year}}{First year in which the author published.}
\item{\code{last_year}}{Most recent year in which the author published.}
}
}
\source{
Curated by INLPO staff.
}
\usage{
authors
}
\description{
Authors who have contributed to publications by the U.S. Geological
Survey (USGS), Idaho Water Science Center, Idaho National Laboratory Project
Office (INLPO).
}
\examples{
# Subset Jason Fisher's information and display its structure:
author <- authors["jfisher", ]
str(author, max.level = 3, width = 75, strict.width = "cut")

# Print the author's given name:
author$person |> format(include = "given")
}
\seealso{
\code{\link{pubs}} for the publications referenced by \code{pub_id}.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/terms.R
\docType{data}
\name{terms}
\alias{terms}
\title{Term Frequency from INLPO Publications}
\format{
An object of class 'term' that inherits from the 'data.frame' class and
includes the following columns:
\describe{
\item{\code{term}}{Term (a single word, or a multi-word n-gram), stored as a factor.
Terms are lowercase ASCII.}
\item{\code{pub_id}}{Identifier of the publication in which the term appears, stored
as a factor; a foreign key referencing the \code{pub_id} primary key of the
\code{\link{pubs}} table.}
\item{\code{freq}}{Number of times the term occurs in that publication.}
}
}
\source{
The publication text was sourced from the original PDF documents using the
\code{\link{get_pdf_text}} function, and term frequencies were extracted from the text
using the \code{\link{mine_text}} function.
}
\usage{
terms
}
\description{
A term-frequency table giving the number of times each term appears in
each publication, for publications by the U.S. Geological Survey (USGS), Idaho
Water Science Center, Idaho National Laboratory Project Office (INLPO). The table
has one row per term-publication pairing.
}
\examples{
str(terms, max.level = 3, width = 75, strict.width = "cut")

# Most frequent terms across all publications:
freq_by_term <- tapply(terms$freq, terms$term, FUN = sum)
sort(freq_by_term, decreasing = TRUE) |> head(n = 10)
}
\seealso{
\code{\link{mine_text}} to build a term-frequency table, \code{\link{search_terms}} to search
within it, and \code{\link{make_wordcloud}} to visualize it.
}
\keyword{datasets}
